import importlib
import sys
from inspect import getmembers, isfunction
from pathlib import Path

from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.simulation.implants.Signal import Signal
from immuneML.util.Logger import log
from immuneML.util.ParameterValidator import ParameterValidator


class SignalParser:
    keyword = "signals"
    custom_func_keys = ['source_file', 'is_present_func']

    @staticmethod
    @log
    def parse(signals: dict, symbol_table: SymbolTable):
        for key, signal_spec in signals.items():

            assert "__" not in key, f"{SignalParser.__name__}: '__' is not valid part of signal names, please rename the signal."

            if "motifs" in signal_spec:
                signal = _parse_signal_with_motifs(key, signal_spec, symbol_table)
            else:
                signal = _parse_custom_func_signal(key, signal_spec)

            symbol_table.add(key, SymbolType.SIGNAL, signal)

        return symbol_table, signals


def _parse_custom_func_signal(key: str, signal_spec: dict) -> Signal:
    assert all(k in signal_spec for k in SignalParser.custom_func_keys), \
        f"Signal {key}: for signals with custom functions the following keys need to be defined: " \
        f"{SignalParser.custom_func_keys}, got: {list(signal_spec.keys())}"

    assert signal_spec['source_file'].endswith(".py") and Path(signal_spec['source_file']).is_file(), \
        f"Signal {key}: no file {signal_spec['source_file']}"

    sys.path.insert(0, str(Path(signal_spec['source_file']).parent))
    source_file = importlib.import_module(Path(signal_spec['source_file']).name[:-3])
    functions = getmembers(source_file, isfunction)
    is_present_func = [func for func in functions if signal_spec['is_present_func'] == func[0]]

    assert len(is_present_func) == 1, \
        f"Signal {key}: no function named {signal_spec['is_present_func']}."

    return Signal(id=key, is_present_custom_func=is_present_func[0][1])


def _parse_signal_with_motifs(key: str, signal_spec: dict, symbol_table: SymbolTable) -> Signal:
    assert all(k not in signal_spec for k in SignalParser.custom_func_keys) or \
           all(signal_spec[k] is None for k in SignalParser.custom_func_keys), \
        f"Signal {key}: define either motifs or the custom function."

    valid_motif_keys = symbol_table.get_keys_by_type(SymbolType.MOTIF)
    signal_motifs = []

    for motif_group in signal_spec['motifs']:
        if isinstance(motif_group, str):
            ParameterValidator.assert_in_valid_list(motif_group, valid_motif_keys, SignalParser.__name__,
                                                    f'{key}:motifs')
            signal_motifs.append(symbol_table.get(motif_group))
        elif isinstance(motif_group, list):
            assert len(
                motif_group) == 2, f"{SignalParser.__name__}: {len(motif_group)} motifs specified for signal {key}, but only 2 allowed."
            for motif in motif_group:
                ParameterValidator.assert_in_valid_list(motif, valid_motif_keys, SignalParser.__name__, f'{key}:motifs')
            signal_motifs.append([symbol_table.get(motif_id) for motif_id in motif_group])

    check_clonal_frequency(signal_spec)

    signal = Signal(key, signal_motifs, v_call=signal_spec.get('v_call'), j_call=signal_spec.get('j_call'),
                    clonal_frequency=signal_spec.get('clonal_frequency', None),
                    sequence_position_weights=signal_spec.get('sequence_position_weights', {}))
    return signal


def check_clonal_frequency(spec: dict, name: str = 'clonal_frequency', location: str = SignalParser.__name__):
    if name in spec and spec[name] is not None:
        ParameterValidator.assert_type_and_value(spec[name], dict, location, name)
        ParameterValidator.assert_all_in_valid_list(spec[name].keys(), ['a', 'loc'], location, name)



import copy
import logging
import os
import subprocess
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.figure_factory as ff
import plotly.graph_objects as go
from scipy.cluster.hierarchy import linkage, fcluster

from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class CompAIRRClusteringReport(DataReport):
    """
    A report that uses CompAIRR to compute repertoire distances based on sequence overlap and performs hierarchical
    clustering on the resulting distance matrix. The clustering results are visualized using a dendrogram,
    colored by a specified label.

    The report assumes that CompAIRR (https://github.com/uio-bmi/compairr) has been installed.

    .. note::

        This is an experimental feature.

    **Specification arguments**:

    - labels (list): The list of labels to highlight below the dendrogram. The labels should be present in the dataset.

    - compairr_path (str): Path to the CompAIRR executable.

    - indels (bool): Whether to allow insertions/deletions when matching sequences (default: False)

    - ignore_counts (bool): Whether to ignore sequence counts when computing overlap (default: False)

    - ignore_genes (bool): Whether to ignore V/J gene assignments when matching sequences (default: False)

    - threads (int): Number of threads to use for CompAIRR computation (default: 4)

    - linkage_method (str): The linkage method to use for hierarchical clustering (default: 'single')

    - is_cdr3 (bool): Whether the sequences represent CDR3s (default: True)

    - clustering_criterion (str): The criterion to use for clustering (default: 'distance'), as defined in
      scipy.cluster.hierarchy.linkage; valid values are 'distance', 'maxclust', 'monocrit', 'maxclust_monocrit'

    - clustering_threshold (float): The threshold for the clustering algorithm (default: 0.5), mapped to 't' parameter
      in scipy.cluster.hierarchy.fcluster

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_compairr_clustering_report:
                    CompAIRRClusteringReport:
                        labels: [disease]
                        compairr_path: /path/to/compairr
                        indels: false
                        ignore_counts: true
                        ignore_genes: true
                        threads: 4
                        linkage_method: single
                        is_cdr3: true
                        clustering_criterion: distance
                        clustering_threshold: 0.5

    """

    @classmethod
    def build_object(cls, **kwargs):
        valid_keys = ['labels', 'compairr_path', 'indels', 'ignore_counts', 'ignore_genes', 'threads', 'linkage_method',
                      'is_cdr3', 'clustering_threshold', 'clustering_criterion', 'name']
        ParameterValidator.assert_keys(kwargs, valid_keys, CompAIRRClusteringReport.__name__,
                                       CompAIRRClusteringReport.__name__, True)
        ParameterValidator.assert_type_and_value(kwargs['labels'], list, CompAIRRClusteringReport.__name__, 'labels')
        ParameterValidator.assert_in_valid_list(kwargs['clustering_criterion'],
                                                ['distance', 'maxclust', 'monocrit', 'maxclust_monocrit'],
                                                CompAIRRClusteringReport.__name__, 'clustering_criterion')
        assert isinstance(kwargs['clustering_threshold'], (int, float)), 'clustering_threshold must be a number'

        return CompAIRRClusteringReport(**kwargs)

    def __init__(self, dataset: RepertoireDataset = None, result_path: Path = None, labels: List[str] = None,
                 compairr_path: str = None, indels: bool = False, ignore_counts: bool = False,
                 ignore_genes: bool = False, threads: int = 4, linkage_method: str = 'single', is_cdr3: bool = True,
                 name: str = None, clustering_threshold: float = 0.5, clustering_criterion: str = 'distance'):
        super().__init__(dataset=dataset, result_path=result_path, name=name)
        self.labels = labels
        self.linkage_method = linkage_method
        self.compairr_path = compairr_path
        self.indels = indels
        self.ignore_counts = ignore_counts
        self.ignore_genes = ignore_genes
        self.threads = threads
        self.is_cdr3 = is_cdr3
        self.clustering_threshold = clustering_threshold
        self.clustering_criterion = clustering_criterion

    def check_prerequisites(self) -> bool:
        if not self.compairr_path:
            logging.warning("CompAIRR path not provided. CompAIRR must be installed and available in the system PATH.")
            return False
        if not isinstance(self.dataset, RepertoireDataset):
            logging.warning("CompAIRRClusteringReport requires a RepertoireDataset as input.")
            return False
        return True

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        self.check_label()

        similarity_matrix = self.compare_repertoires()
        distance_matrix = self.compute_distance_matrix(similarity_matrix)

        linkage_matrix = linkage(distance_matrix.values, method='single')
        predictions = fcluster(linkage_matrix, self.clustering_threshold, criterion=self.clustering_criterion)

        metadata = self.dataset.get_metadata(self.labels + ['subject_id'], return_df=True)

        # Save outputs
        distance_output = self._store_matrix(distance_matrix, 'distance_matrix')
        similarity_output = self._store_matrix(similarity_matrix, 'similarity_matrix')
        similarity_heatmap = self._make_similarity_heatmap(similarity_matrix, 'similarity')
        dendrogram_output = self._create_dendrogram(distance_matrix, metadata, linkage_matrix)
        cluster_output = self._store_cluster_assignments(self.dataset.get_example_ids(), predictions)

        return ReportResult(
            name=self.name,
            info="Hierarchical clustering of repertoires based on CompAIRR sequence overlap",
            output_figures=[dendrogram_output, similarity_heatmap],
            output_tables=[distance_output, similarity_output, cluster_output]
        )

    def compute_distance_matrix(self, similarity_matrix) -> pd.DataFrame:
        return pd.DataFrame(data=1 - similarity_matrix.values, columns=similarity_matrix.columns,
                            index=similarity_matrix.index)

    def compare_repertoires(self) -> pd.DataFrame:
        # Create similarity matrix
        n_repertoires = len(self.dataset.repertoires)
        similarity_matrix = self._init_similarity_matrix(n_repertoires)

        # Compare repertoires pairwise
        for i in range(n_repertoires):
            rep1 = self.dataset.repertoires[i]
            tmp_rep_file1 = self.result_path / f"{rep1.identifier}_tmp.tsv"
            self._make_clean_rep_files(rep1, tmp_rep_file1)

            similarity_matrix.loc[rep1.identifier, rep1.identifier] = 1
            for j in range(i + 1, n_repertoires):
                rep2 = self.dataset.repertoires[j]

                tmp_rep_file2 = self.result_path / f"{rep2.identifier}_tmp.tsv"
                self._make_clean_rep_files(rep2, tmp_rep_file2)

                similarity_matrix = self._run_compairr(rep1, rep2, tmp_rep_file1, tmp_rep_file2, similarity_matrix)

                os.remove(str(tmp_rep_file2))

            os.remove(str(tmp_rep_file1))

            logging.info(f"CompAIRRClusteringReport: Finished processing repertoire "
                         f"{rep1.identifier} ({i + 1}/{n_repertoires})")

        return similarity_matrix

    def check_label(self):
        if any(label not in self.dataset.get_label_names() for label in self.labels):
            raise ValueError(
                f"Labels '{self.labels}' not found in dataset metadata. Available labels: {self.dataset.get_label_names()}")

    def _init_similarity_matrix(self, n_repertoires: int) -> pd.DataFrame:
        return pd.DataFrame(
            np.zeros((n_repertoires, n_repertoires)),
            index=[rep.identifier for rep in self.dataset.repertoires],
            columns=[rep.identifier for rep in self.dataset.repertoires]
        )

    def _make_clean_rep_files(self, rep: Repertoire, tmp_file_path: Path):
        df = pd.read_csv(rep.data_filename, sep="\t")[['cdr3_aa', 'v_call', 'j_call', 'duplicate_count']]
        if (df['duplicate_count'] == -1).any() or df['duplicate_count'].isna().any():
            logging.warning(f"CompAIRRClusteringReport: Duplicate count -1 or NA found in repertoire {rep.identifier}. "
                            "Setting to 1, but it should be rechecked.")
            df['duplicate_count'] = df['duplicate_count'].replace(-1, 1)
            df['duplicate_count'] = df['duplicate_count'].replace(pd.NA, 1)

        df.to_csv(tmp_file_path, sep="\t", index=False)

    def _run_compairr(self, rep1: Repertoire, rep2: Repertoire, tmp_rep_file1: Path, tmp_rep_file2: Path,
                      similarity_matrix: pd.DataFrame):
        output_file = self.result_path / f"overlap_{rep1.identifier}_{rep2.identifier}.tsv"

        cmd_args = [str(self.compairr_path), "--matrix", str(tmp_rep_file1), str(tmp_rep_file2),
                    "-o", str(output_file)]
        if self.indels:
            cmd_args.append("-i")
        if self.ignore_counts:
            cmd_args.append("--ignore-counts")
        if self.ignore_genes:
            cmd_args.append("--ignore-genes")
        if self.is_cdr3:
            cmd_args.append("--cdr3")
        cmd_args.extend(["-t", str(self.threads)])
        cmd_args.extend(["-s", 'Jaccard'])

        result = subprocess.run(cmd_args, capture_output=True, text=True)
        if result.returncode != 0:
            raise RuntimeError(
                f"CompAIRRClusteringReport: Error running CompAIRR for repertoires {rep1.identifier} and "
                f"{rep2.identifier}:\n{result.stderr}")

        elif output_file.is_file():
            similarity = pd.read_csv(output_file, sep="\t").iloc[0, 1]
            similarity_matrix.loc[rep1.identifier, rep2.identifier] = similarity
            similarity_matrix.loc[rep2.identifier, rep1.identifier] = similarity

            os.remove(str(output_file))

        return similarity_matrix

    def _store_matrix(self, matrix: pd.DataFrame, name: str) -> ReportOutput:
        output_path = self.result_path / f"{name}.tsv"
        matrix.to_csv(output_path, sep="\t")
        return ReportOutput(path=output_path, name=name)

    def _make_similarity_heatmap(self, matrix: pd.DataFrame, name: str) -> ReportOutput:

        metadata = self.dataset.get_metadata(field_names=None, return_df=True)

        subject_ids = metadata['subject_id'].tolist() if 'subject_id' in metadata.columns \
            else [str(el) for el in range(matrix.shape[0])]

        mask = np.zeros_like(matrix, dtype=bool)
        mask[np.triu_indices_from(mask)] = True

        matrix_values = copy.deepcopy(matrix.values)
        matrix_values[mask] = np.nan
        matrix_text = matrix_values.round(3).astype(str)
        matrix_text = np.where(np.isnan(matrix_values), '', matrix_text)

        fig = go.Figure(data=go.Heatmap(z=matrix_values, x=subject_ids[:-1] + [""], y=[""] + subject_ids[1:],
                                        colorscale='Darkmint',
                                        hoverongaps=False, text=matrix_text, texttemplate="%{text}"))
        fig.update_layout(template="plotly_white", title=f"{name} matrix", xaxis_title="repertoire",
                          yaxis_title='repertoire')
        fig.update_xaxes(showgrid=False)
        fig.update_yaxes(showgrid=False, autorange='reversed')
        file_path = PlotlyUtil.write_image_to_file(fig, self.result_path / f"{name}_heatmap.html", matrix_text.shape[0])

        return ReportOutput(path=file_path, name=f"{name} heatmap")

    def _store_cluster_assignments(self, repertoire_ids, clusters) -> ReportOutput:
        output_path = self.result_path / "cluster_assignments.tsv"
        cluster_df_data = {
            'repertoire_id': repertoire_ids,
            'cluster': clusters
        }

        metadata = self.dataset.get_metadata(field_names=None, return_df=True)
        if 'subject_id' in metadata.columns:
            cluster_df_data['subject_id'] = metadata['subject_id'].tolist()

        cluster_df = pd.DataFrame(cluster_df_data)
        cluster_df.to_csv(output_path, sep="\t", index=False)
        return ReportOutput(
            path=output_path,
            name="Cluster Assignments",
        )

    def _create_dendrogram(self, distance_matrix, metadata: pd.DataFrame, linkage_matrix) -> ReportOutput:
        """
        Parameters:
        -----------
        repertoire_ids : list
            List of repertoire IDs
        distance_matrix : DataFrame
            Distance matrix
        external_labels : dict
            Dictionary mapping repertoire_ids to their labels/categories
        """
        output_path = self.result_path / "dendrogram.html"
        repertoire_ids = metadata['subject_id'].tolist() if 'subject_id' in metadata.columns else self.dataset.get_example_ids()
        metadata['subject_id'] = repertoire_ids

        # Create dummy data array for dendrogram
        n = distance_matrix.shape[0]
        dummy_data = np.zeros((n, 2))  # 2D dummy data

        # Initialize figure by creating upper dendrogram only
        fig = ff.create_dendrogram(dummy_data, orientation='bottom',
                                   labels=repertoire_ids, linkagefun=lambda x: linkage_matrix,
                                   distfun=lambda x: distance_matrix.values)

        # Remove trace numbers from dendrogram hover
        for i in range(len(fig['data'])):
            fig['data'][i]['yaxis'] = 'y2'
            fig['data'][i]['hovertemplate'] = None
            fig['data'][i]['showlegend'] = False

        # Create a mapping from repertoire ids to indices
        label_to_idx = {label: idx for idx, label in enumerate(repertoire_ids)}

        # Get the ordering of leaves from the top dendrogram and convert to indices
        dendro_leaves = [label_to_idx[label] for label in fig['layout']['xaxis']['ticktext']]
        ordered_ids = [repertoire_ids[idx] for idx in dendro_leaves]

        metadata = metadata.set_index('subject_id')

        annotation_start = 0.72
        annotation_end = 0.82
        annotation_gap = 0
        annotation_height = (annotation_end - annotation_start - (len(self.labels) - 1) * annotation_gap) / len(self.labels)

        for i, label in enumerate(self.labels):
            yaxis_name = f'y{i + 3}'
            y_domain = [
                annotation_start + i * (annotation_height + annotation_gap),
                annotation_start + i * (annotation_height + annotation_gap) + annotation_height
            ]

            annotation_values = [metadata.loc[ind, label] for ind in ordered_ids]
            unique_labels = sorted(metadata[label].unique().tolist())
            label_to_num = {label: j for j, label in enumerate(unique_labels)}
            annotation_numeric = [label_to_num[v] for v in annotation_values]

            # Add annotation heatmap
            fig.add_trace(go.Heatmap(
                x=fig['layout']['xaxis']['tickvals'],
                y=[label],
                z=np.array(annotation_numeric).reshape(1, -1),
                showlegend=False,
                showscale=False,
                colorscale=px.colors.qualitative.Plotly[1:],
                hovertemplate=label + ': %{text}<br>Repertoire: %{x}<extra></extra>',
                text=[annotation_values],
                yaxis=yaxis_name
            ))

            # Update layout for this axis
            fig.update_layout(**{
                f'yaxis{i + 3}': dict(
                    domain=y_domain,
                    mirror=False,
                    showgrid=False,
                    showline=False,
                    zeroline=False
                )
            })

        # Reorder the distance matrix according to dendrogram leaves
        heat_data = distance_matrix.values[dendro_leaves, :]
        heat_data = heat_data[:, dendro_leaves]

        # Create Heatmap with custom hover template
        heatmap = [
            go.Heatmap(
                x=fig['layout']['xaxis']['tickvals'],
                y=fig['layout']['xaxis']['tickvals'],
                z=heat_data,
                colorscale=px.colors.sequential.Blues,
                showlegend=False,
                showscale=False,
                hovertemplate='X: %{x}<br>Y: %{y}<br>Distance: %{z}<extra></extra>'
            )
        ]

        # Add Heatmap Data to Figure
        for data in heatmap:
            fig.add_trace(data)

        # Edit Layout with adjusted domains for the new annotation row
        fig.update_layout({
            'height': 700 + len(self.labels) * 20,  # Adjust height based on number of labels
            'autosize': True,
            'showlegend': False,
            'hovermode': 'closest',
            'template': 'plotly_white',
            'xaxis': {
                'domain': [0, 1],
                'mirror': False,
                'showgrid': False,
                'showline': False,
                'zeroline': False,
                'ticks': ""
            },
            'yaxis': {
                'domain': [0, 0.7],  # Reduced height for main heatmap
                'mirror': False,
                'showgrid': False,
                'showline': False,
                'zeroline': False,
                'showticklabels': False,
                'ticks': ""
            },
            'yaxis2': {
                'domain': [0.85, 1],  # Top dendrogram
                'mirror': False,
                'showgrid': False,
                'showline': False,
                'zeroline': False,
                'showticklabels': False,
                'ticks': ""
            }
        })

        output_path = PlotlyUtil.write_image_to_file(fig, output_path, metadata.shape[0])

        return ReportOutput(path=output_path, name="hierarchical_clustering_of_repertoires")



import copy
from pathlib import Path
from typing import List

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.Label import Label
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.example_weighting.ExampleWeightingStrategy import ExampleWeightingStrategy
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.core.HPUtil import HPUtil
from immuneML.hyperparameter_optimization.states.HPItem import HPItem
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.reports.ReportUtil import ReportUtil
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.steps.MLMethodTrainer import MLMethodTrainer
from immuneML.workflows.steps.MLMethodTrainerParams import MLMethodTrainerParams


class MLProcess:
    """
    Class that implements the machine learning process:
        1. encodes the training dataset
        2. encodes the test dataset (using parameters learnt in step 1 if there are any such parameters)
        3. trains the ML method on encoded training dataset
        4. assesses the method's performance on encoded test dataset

    It performs the task for a given label configuration, and given list of metrics (used only in the assessment step).
    """

    def __init__(self, train_dataset: Dataset, test_dataset: Dataset, label: Label, metrics: set,
                 optimization_metric: ClassificationMetric, path: Path, ml_reports: List[MLReport] = None,
                 encoding_reports: list = None, data_reports: list = None, number_of_processes: int = 2,
                 label_config: LabelConfiguration = None, report_context: dict = None, hp_setting: HPSetting = None,
                 example_weighting: ExampleWeightingStrategy = None,
                 sequence_type: SequenceType = SequenceType.AMINO_ACID, region_type: RegionType = RegionType.IMGT_CDR3):
        self.train_dataset = train_dataset
        self.test_dataset = test_dataset
        self.label = label
        self.label_config = label_config
        self.method = copy.deepcopy(hp_setting.ml_method)
        self.path = PathBuilder.build(path) if path is not None else None
        self.ml_settings_export_path = path / "ml_settings_config" if path is not None else None
        self.ml_score_path = path / "ml_score.csv" if path is not None else None
        self.train_predictions_path = path / "train_predictions.csv" if path is not None else None
        self.test_predictions_path = path / "test_predictions.csv" if path is not None else None
        self.report_path = PathBuilder.build(path / "reports") if path is not None else None
        self.number_of_processes = number_of_processes
        assert all([isinstance(metric, ClassificationMetric) for metric in metrics]), \
            "MLProcess: metrics are not set to be an instance of Metric."
        self.metrics = metrics
        self.optimization_metric = optimization_metric
        self.ml_reports = ml_reports if ml_reports is not None else []
        self.encoding_reports = encoding_reports if encoding_reports is not None else []
        self.data_reports = data_reports if data_reports is not None else []
        self.report_context = report_context
        self.hp_setting = copy.deepcopy(hp_setting)
        self.example_weighting = example_weighting
        self.sequence_type = sequence_type
        self.region_type = region_type

    def _set_paths(self):
        if self.path is None:
            raise RuntimeError("MLProcess: path is not set, stopping execution...")
        self.ml_settings_export_path = self.path / "ml_settings_config"
        self.ml_score_path = self.path / "ml_score.csv"
        self.train_predictions_path = self.path / "train_predictions.csv"
        self.test_predictions_path = self.path / "test_predictions.csv"
        self.report_path = PathBuilder.build(self.path / "reports")

    def run(self, split_index: int) -> HPItem:

        print_log(f"Evaluating hyperparameter setting: {self.hp_setting}...", include_datetime=True)

        PathBuilder.build(self.path)
        self._set_paths()

        processed_dataset = HPUtil.preprocess_dataset(self.train_dataset, self.hp_setting.preproc_sequence,
                                                      self.path / "preprocessed_train_dataset",
                                                      self.report_context,
                                                      number_of_processes=self.number_of_processes)

        weighted_dataset = HPUtil.weight_examples(dataset=processed_dataset, weighting_strategy=self.example_weighting,
                                                  path=self.path / "weighted_train_datasets",
                                                  learn_model=True, number_of_processes=self.number_of_processes)

        encoded_train_dataset = HPUtil.encode_dataset(weighted_dataset, self.hp_setting, self.path / "encoded_datasets",
                                                      learn_model=True, context=self.report_context,
                                                      number_of_processes=self.number_of_processes,
                                                      label_configuration=self.label_config,
                                                      sequence_type=self.sequence_type,
                                                      region_type=self.region_type)

        method = self._train_method(encoded_train_dataset)

        encoding_train_results = ReportUtil.run_encoding_reports(encoded_train_dataset, self.encoding_reports,
                                                                 self.report_path / "encoding_train",
                                                                 self.number_of_processes)

        hp_item = self._assess_on_test_dataset(encoded_train_dataset, encoding_train_results, method, split_index)

        print_log(f"Completed hyperparameter setting {self.hp_setting}.\n", include_datetime=True)

        return hp_item

    def _train_method(self, dataset) -> MLMethod:
        method = MLMethodTrainer.run(MLMethodTrainerParams(
            method=copy.deepcopy(self.hp_setting.ml_method),
            result_path=self.path / "ml_method",
            dataset=dataset,
            label=self.label,
            train_predictions_path=self.train_predictions_path,
            model_selection_cv=self.hp_setting.ml_params["model_selection_cv"],
            model_selection_n_folds=self.hp_setting.ml_params["model_selection_n_folds"],
            cores_for_training=self.number_of_processes,
            optimization_metric=self.optimization_metric.name.lower()
        ))
        return method

    def _assess_on_test_dataset(self, encoded_train_dataset, encoding_train_results, method, split_index) -> HPItem:
        if self.test_dataset is not None and self.test_dataset.get_example_count() > 0:
            processed_test_dataset = HPUtil.preprocess_dataset(self.test_dataset, self.hp_setting.preproc_sequence,
                                                               self.path / "preprocessed_test_dataset")

            weighted_test_dataset = HPUtil.weight_examples(dataset=processed_test_dataset,
                                                           weighting_strategy=self.example_weighting,
                                                           path=self.path / "weighted_test_datasets", learn_model=False,
                                                           number_of_processes=self.number_of_processes)

            encoded_test_dataset = HPUtil.encode_dataset(weighted_test_dataset, self.hp_setting,
                                                         self.path / "encoded_datasets",
                                                         learn_model=False, context=self.report_context,
                                                         number_of_processes=self.number_of_processes,
                                                         label_configuration=self.label_config,
                                                         sequence_type=self.sequence_type,
                                                         region_type=self.region_type)

            performance = HPUtil.assess_performance(method, self.metrics, self.optimization_metric,
                                                    encoded_test_dataset, split_index, self.path,
                                                    self.test_predictions_path, self.label, self.ml_score_path)

            encoding_test_results = ReportUtil.run_encoding_reports(encoded_test_dataset, self.encoding_reports,
                                                                    self.report_path / "encoding_test",
                                                                    self.number_of_processes)

            model_report_results = ReportUtil.run_ML_reports(encoded_train_dataset, encoded_test_dataset, method,
                                                             self.ml_reports,
                                                             self.report_path / "ml_method", self.hp_setting,
                                                             self.label, self.number_of_processes, self.report_context)

            hp_item = HPItem(method=method, hp_setting=self.hp_setting,
                             train_predictions_path=self.train_predictions_path,
                             test_predictions_path=self.test_predictions_path, train_dataset=self.train_dataset,
                             test_dataset=self.test_dataset, split_index=split_index,
                             model_report_results=model_report_results,
                             encoding_train_results=encoding_train_results, encoding_test_results=encoding_test_results,
                             performance=performance,
                             encoder=self.hp_setting.encoder, ml_settings_export_path=self.ml_settings_export_path)
        else:
            hp_item = HPItem(method=method, hp_setting=self.hp_setting,
                             train_predictions_path=self.train_predictions_path,
                             test_predictions_path=None, train_dataset=self.train_dataset,
                             split_index=split_index, encoding_train_results=encoding_train_results,
                             encoder=self.hp_setting.encoder,
                             ml_settings_export_path=self.ml_settings_export_path)

        return hp_item



# coding=utf-8
# Copyright 2021 The EleutherAI and HuggingFace Teams. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified configuration implementation based on https://github.com/huggingface/transformers/blob/main/src/transformers/models/gptj/configuration_gptj.py
# Modified for ImmuneML. Original code from https://github.com/salesforce/progen

from transformers.configuration_utils import PretrainedConfig
from transformers.utils import logging

logger = logging.get_logger(__name__)


class ProGenConfig(PretrainedConfig):
    model_type = "progen"

    def __init__(
        self,
        vocab_size=50400,
        n_positions=2048,
        n_ctx=2048,
        n_embd=4096,
        n_layer=28,
        n_head=16,
        rotary_dim=64,
        n_inner=None,
        activation_function="gelu_new",
        resid_pdrop=0.0,
        embd_pdrop=0.0,
        attn_pdrop=0.0,
        layer_norm_epsilon=1e-5,
        initializer_range=0.02,
        scale_attn_weights=True,
        gradient_checkpointing=False,
        use_cache=True,
        bos_token_id=50256,
        eos_token_id=50256,
        **kwargs
    ):
        super().__init__(bos_token_id=bos_token_id, eos_token_id=eos_token_id, **kwargs)

        self.vocab_size = vocab_size
        self.n_ctx = n_ctx
        self.n_positions = n_positions
        self.n_embd = n_embd
        self.n_layer = n_layer
        self.n_head = n_head
        self.n_inner = n_inner
        self.rotary_dim = rotary_dim
        self.activation_function = activation_function
        self.resid_pdrop = resid_pdrop
        self.embd_pdrop = embd_pdrop
        self.attn_pdrop = attn_pdrop
        self.layer_norm_epsilon = layer_norm_epsilon
        self.initializer_range = initializer_range
        self.gradient_checkpointing = gradient_checkpointing
        self.scale_attn_weights = scale_attn_weights
        self.use_cache = use_cache

        self.bos_token_id = bos_token_id
        self.eos_token_id = eos_token_id

    @property
    def max_position_embeddings(self):
        return self.n_positions

    @property
    def hidden_size(self):
        return self.n_embd

    @property
    def num_attention_heads(self):
        return self.n_head

    @property
    def num_hidden_layers(self):
        return self.n_layer



import os
from pathlib import Path

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.util.Util import Util as MLUtil
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.dataset_generation.DatasetExportState import DatasetExportState


class DatasetExportHTMLBuilder:

    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(state: DatasetExportState) -> Path:
        """
        Function that builds the HTML files based on the Simulation state.
        Arguments:
            state: SimulationState object including all details of the Simulation instruction
        Returns:
             path to the main HTML file (which is located under state.result_path)
        """
        base_path = PathBuilder.build(state.result_path / "../HTML_output/")
        html_map = DatasetExportHTMLBuilder.make_html_map(state, base_path)
        result_file = base_path / f"DatasetExport_{state.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "DatasetExport.html",
                             template_map=html_map, result_path=result_file)

        return result_file

    @staticmethod
    def make_html_map(state: DatasetExportState, base_path: Path) -> dict:
        html_map = {
            "css_style": Util.get_css_content(DatasetExportHTMLBuilder.CSS_PATH),
            "name": state.name,
            'immuneML_version': MLUtil.get_immuneML_version(),
            "full_specs": Util.get_full_specs_path(base_path),
            "logfile": Util.get_logfile_path(base_path),
            "datasets": [
                {
                    **Util.make_dataset_html_map(dataset),
                    **{
                        "preprocessing_sequence": [
                            {
                                "preprocessing_name": preprocessing.__class__.__name__,
                                "preprocessing_params": ", ".join([f"{key}: {value}" for key, value in vars(preprocessing).items()])
                            } for preprocessing in state.preprocessing_sequence
                        ] if state.preprocessing_sequence is not None else [],
                        "show_preprocessing": state.preprocessing_sequence is not None and len(state.preprocessing_sequence) > 0,
                        "formats": [
                            {
                                "format_name": format_name,
                                "dataset_download_link": os.path.relpath(path=Util.make_downloadable_zip(state.result_path, state.paths[dataset.name][format_name]),
                                                                     start=base_path)
                            } for format_name in state.formats
                        ]
                    }
                } for dataset in state.datasets
            ]
        }

        return html_map



from multiprocessing.pool import Pool
from typing import List

import dill
import numpy as np
import pandas as pd

from immuneML.analysis.SequenceMatcher import SequenceMatcher
from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.reference_encoding.MatchedReferenceUtil import MatchedReferenceUtil
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReadsType import ReadsType


class MatchedSequencesEncoder(DatasetEncoder):
    """
    Encodes the dataset based on the matches between a RepertoireDataset and a reference sequence dataset. The feature
    names are derived from the reference sequences: "v_call_sequence_j_call" (e.g., "TRBV12-3_CASSLGTDTQYF_TRBJ2-7"). If
    there are duplicates in the feature names but sequences have different sequence IDs, the sequence ID is appended to
    the feature name to make it unique.

    This encoding can be used in combination with the :ref:`Matches` report.

    When sum_matches and normalize are set to True, this encoder behaves as described in: Yao, Y. et al. ‘T cell receptor repertoire as a potential diagnostic marker for celiac disease’.
    Clinical Immunology Volume 222 (January 2021): 108621. `doi.org/10.1016/j.clim.2020.108621 <https://doi.org/10.1016/j.clim.2020.108621>`_


    **Dataset type:**

    - RepertoireDatasets


    **Specification arguments:**

    - reference (dict): A dictionary describing the reference dataset file. Import should be specified the same way as
      regular dataset import. It is only allowed to import a sequence dataset here (i.e., is_repertoire and paired are
      False by default, and are not allowed to be set to True).

    - max_edit_distance (int): The maximum edit distance between a target sequence (from the repertoire) and the
      reference sequence.

    - reads (:py:mod:`~immuneML.util.ReadsType`): Reads type signify whether the counts of the sequences in the
      repertoire will be taken into account. If :py:mod:`~immuneML.util.ReadsType.UNIQUE`, only unique sequences
      (clonotypes) are counted, and if :py:mod:`~immuneML.util.ReadsType.ALL`, the sequence 'count' value is summed when
      determining the number of matches. The default value for reads is all.

    - sum_matches (bool): When sum_matches is False, the resulting encoded data matrix contains multiple columns with
      the number of matches per reference sequence. When sum_matches is true, all columns are summed together, meaning
      that there is only one aggregated sum of matches per repertoire in the encoded data.
      To use this encoder in combination with the :ref:`Matches` report, sum_matches must be set to False. When
      sum_matches is set to True, this encoder behaves as described by Yao, Y. et al. By default, sum_matches is False.

    - normalize (bool): If True, the sequence matches are divided by the total number of unique sequences in the
      repertoire (when reads = unique) or the total number of reads in the repertoire (when reads = all).

    - output_count_as_feature: if True, the encoded repertoire is represented by the matches, and by the total number
      of sequences (or reads) in the repertoire, as defined by reads parameter above; by default this is False


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_ms_encoding:
                    MatchedSequences:
                        reference:
                            format: VDJDB
                            params:
                                path: path/to/file.txt
                        max_edit_distance: 1

    """

    def __init__(self, max_edit_distance: int, reference: List[ReceptorSequence], reads: ReadsType, sum_matches: bool,
                 normalize: bool, output_count_as_feature: bool = False, name: str = None):
        super().__init__(name=name)
        self.max_edit_distance = max_edit_distance
        self.reference_sequences = reference
        self.reads = reads
        self.sum_matches = sum_matches
        self.normalize = normalize
        self.output_count_as_feature = output_count_as_feature
        self.feature_count = (1 if self.sum_matches else len(self.reference_sequences)) + int(output_count_as_feature)

    @staticmethod
    def _prepare_parameters(max_edit_distance: int, reference: dict, reads: str, sum_matches: bool, normalize: bool,
                            output_count_as_feature: bool = False, name: str = None):
        location = "MatchedSequencesEncoder"

        ParameterValidator.assert_type_and_value(max_edit_distance, int, location, "max_edit_distance", min_inclusive=0)
        ParameterValidator.assert_type_and_value(sum_matches, bool, location, "sum_matches")
        ParameterValidator.assert_type_and_value(output_count_as_feature, bool, location, "normalize")
        ParameterValidator.assert_type_and_value(normalize, bool, location, "output_count_as_feature")
        ParameterValidator.assert_in_valid_list(reads.upper(), [item.name for item in ReadsType], location, "reads")

        if output_count_as_feature and normalize:
            raise RuntimeError(f"{MatchedSequencesEncoder.__name__}: normalize and output_count_as_feature cannot \n"
                               f"be both set to True at the same time. The sequence count (or reads count) can either \n"
                               f"be used for normalization or included as a separate output.")

        reference_sequences = MatchedReferenceUtil.prepare_reference(reference_params=reference, location=location, paired=False)

        return {
            "max_edit_distance": max_edit_distance,
            "reference": reference_sequences,
            "reads": ReadsType[reads.upper()],
            "sum_matches": sum_matches,
            "normalize": normalize,
            "output_count_as_feature": output_count_as_feature,
            "name": name
        }

    @staticmethod
    def build_object(dataset=None, **params):
        if isinstance(dataset, RepertoireDataset):
            prepared_parameters = MatchedSequencesEncoder._prepare_parameters(**params)
            return MatchedSequencesEncoder(**prepared_parameters)
        else:
            raise ValueError("MatchedSequencesEncoder is not defined for dataset types which are not RepertoireDataset.")

    def encode(self, dataset, params: EncoderParams):

        cache_key = CacheHandler.generate_cache_key(self._prepare_caching_params(dataset, params))
        encoded_dataset = CacheHandler.memo(cache_key,
                                            lambda: self._encode_new_dataset(dataset, params))

        return encoded_dataset

    def _prepare_caching_params(self, dataset, params: EncoderParams):

        encoding_params_desc = {"max_edit_distance": self.max_edit_distance,
                                "reference_sequences": sorted([seq.get_sequence() + seq.v_call + seq.j_call
                                                               for seq in self.reference_sequences]),
                                "reads": self.reads.name,
                                "sum_matches": self.sum_matches,
                                "normalize": self.normalize,
                                "output_count_as_feature": self.output_count_as_feature}

        return (("dataset_identifiers", tuple(dataset.get_example_ids())),
                ("dataset_metadata", dataset.metadata_file),
                ("dataset_type", dataset.__class__.__name__),
                ("labels", tuple(params.label_config.get_labels_by_name())),
                ("encoding", MatchedSequencesEncoder.__name__),
                ("learn_model", params.learn_model),
                ("encoding_params", encoding_params_desc),)

    def _encode_new_dataset(self, dataset, params: EncoderParams):
        encoded_repertoires, labels = self._encode_repertoires(dataset, params)

        encoded_repertoires = self._normalize(dataset, encoded_repertoires) if self.normalize else encoded_repertoires

        feature_annotations = None if self.sum_matches else self._get_feature_info()
        if self.sum_matches:
            feature_names = [f"sum_of_{self.reads.value}_reads"]
            if self.output_count_as_feature:
                feature_names += ['sequence_count_in_repertoire']
        else:
            feature_names = list(feature_annotations["sequence_desc"])

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(
            examples=encoded_repertoires,
            labels=labels,
            feature_names=feature_names,
            feature_annotations=feature_annotations,
            example_ids=[repertoire.identifier for repertoire in dataset.get_data()],
            encoding=MatchedSequencesEncoder.__name__,
            info={'sequence_type': params.sequence_type,
                  'region_type': params.region_type}
        )

        return encoded_dataset

    def _normalize(self, dataset, encoded_repertoires):
        if self.reads == ReadsType.UNIQUE:
            repertoire_totals = np.asarray([[repertoire.get_element_count() for repertoire in dataset.get_data()]]).T
        else:
            repertoire_totals = np.asarray([[sum(repertoire.data.duplicate_count) for repertoire in dataset.get_data()]]).T

        return encoded_repertoires / repertoire_totals

    def _get_feature_info(self):
        """
        returns a pandas dataframe containing:
         - sequence id
         - chain
         - amino acid sequence
         - v call
         - j call
        """

        features = [[] for i in range(0, len(self.reference_sequences))]

        for i, sequence in enumerate(self.reference_sequences):
            features[i] = [sequence.sequence_id,
                           sequence.locus,
                           sequence.sequence_aa,
                           sequence.v_call,
                           sequence.j_call,
                           self._get_sequence_desc(sequence)]

        if self.output_count_as_feature:
            features += [['', '', '', '', '', 'sequence_count_in_repertoire']]

        features = pd.DataFrame(features, columns=["sequence_id", "locus", "sequence", "v_call", "j_call", "sequence_desc"])
        if features['sequence_desc'].unique().shape[0] < features.shape[0]:
            features.loc[:, 'sequence_desc'] = [row['sequence_desc'] + "_" + row['sequence_id'] for ind, row in features.iterrows()]

        return features

    def _get_sequence_desc(self, sequence: ReceptorSequence) -> str:
        desc = ""
        if sequence.v_call not in [None, ""]:
            desc += f"{sequence.v_call}_"

        desc += sequence.sequence_aa if sequence.sequence_aa != "" else sequence.sequence

        if sequence.j_call not in ["", None]:
            desc += f"_{sequence.j_call}"

        return desc

    def _encode_repertoires(self, dataset: RepertoireDataset, params):
        labels = {label: [] for label in params.label_config.get_labels_by_name()} if params.encode_labels else None

        with Pool(params.pool_size) as pool:
            encoded_repertories = np.array(pool.map(self._get_repertoire_matches_to_reference,
                                                    [dill.dumps(rep) for rep in dataset.repertoires]))

        for repertoire in dataset.repertoires:
            for label_name in params.label_config.get_labels_by_name():
                labels[label_name].append(repertoire.metadata[label_name])

        return encoded_repertories, labels

    def _get_repertoire_matches_to_reference(self, repertoire):

        if isinstance(repertoire, bytes):
            repertoire = dill.loads(repertoire)

        return CacheHandler.memo_by_params(
            (("repertoire_identifier", repertoire.identifier),
             ("encoding", MatchedSequencesEncoder.__name__),
             ("readstype", self.reads.name),
             ("sum_matches", self.sum_matches),
             ("max_edit_distance", self.max_edit_distance),
             ("reference_sequences", tuple(
                 [(seq.locus, seq.sequence_aa, seq.v_call, seq.j_call) for seq in self.reference_sequences]))),
            lambda: self._compute_matches_to_reference(repertoire))

    def _compute_matches_to_reference(self, repertoire: Repertoire):
        matcher = SequenceMatcher()
        matches = np.zeros(self.feature_count, dtype=int)
        rep_seqs = repertoire.sequences()

        for i, reference_seq in enumerate(self.reference_sequences):

            for repertoire_seq in rep_seqs:
                if matcher.matches_sequence(reference_seq, repertoire_seq, max_distance=self.max_edit_distance):
                    matches_idx = 0 if self.sum_matches else i
                    match_count = 1 if self.reads == ReadsType.UNIQUE else repertoire_seq.duplicate_count
                    matches[matches_idx] += match_count

        if self.output_count_as_feature:
            duplicate_counts = repertoire.data.duplicate_count
            matches[-1] = sum(duplicate_counts) if self.reads == ReadsType.ALL else len(duplicate_counts)

        return matches



import logging
from pathlib import Path

import pandas as pd
import plotly.graph_objects as go
import plotly.express as px

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class LabelOverlap(DataReport):
    """
    This report creates a heatmap where the columns are the values of one label and rows are the values of another label,
    and the cells contain the number of samples that have both label values. It works for any dataset type.

    **Specification arguments:**

    - column_label (str): Name of the label to be used as columns in the heatmap.

    - row_label (str): Name of the label to be used as rows in the heatmap.

    **YAML specification:**

    .. code-block:: yaml

        my_data_report:
            LabelOverlap:
                column_label: epitope
                row_label: batch

    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None, name: str = None,
                 number_of_processes: int = 1,
                 column_label: str = None, row_label: str = None):
        super().__init__(dataset, result_path, name, number_of_processes)
        self.column_label = column_label
        self.row_label = row_label

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()), ["column_label", "row_label"], "LabelOverlap",
                                               'LabelOverlap')
        ParameterValidator.assert_type_and_value(kwargs["column_label"], str, "LabelOverlap", "column_label")
        ParameterValidator.assert_type_and_value(kwargs["row_label"], str, "LabelOverlap", "row_label")
        return LabelOverlap(column_label=kwargs["column_label"], row_label=kwargs["row_label"], name=kwargs.get("name", "LabelOverlap"))

    def check_prerequisites(self):
        if self.column_label not in self.dataset.get_label_names() or self.row_label not in self.dataset.get_label_names():
            logging.warning(f"One or both of the specified labels ({[self.column_label, self.row_label]}) do not exist in the dataset.")
            return False
        return True

    def _generate(self) -> ReportResult:
        # Get metadata for both labels
        metadata = self.dataset.get_metadata([self.column_label, self.row_label])

        # Create a cross-tabulation of the two labels
        overlap_matrix = pd.crosstab(metadata[self.row_label], metadata[self.column_label])

        # Save as CSV
        PathBuilder.build(self.result_path)
        csv_path = self.result_path / 'label_overlap.csv'
        overlap_matrix.to_csv(csv_path)

        # Create heatmap using plotly
        fig = go.Figure(data=go.Heatmap(
            z=overlap_matrix.values,
            x=overlap_matrix.columns,
            y=overlap_matrix.index,
            colorscale=px.colors.sequential.Viridis,  # Custom teal colorscale
            text=overlap_matrix.values,
            texttemplate="%{text}",
            hovertemplate=f"{self.row_label}: " + "%{y}<br>" + f"{self.column_label}: "
                          + "%{x}<br>Count: %{z}<extra></extra>",
            textfont={"size": 14},
            showscale=False,  # Hide the color scale legend
            hoverongaps=False,
        ))

        # Update layout for better readability
        fig.update_layout(
            title=f"Label Overlap: {self.row_label} vs {self.column_label}",
            xaxis_title=self.column_label,
            yaxis_title=self.row_label,
            template="plotly_white",
            font=dict(size=12),
            height=max(600, 40 * len(overlap_matrix.index) + 200),
        )

        # Save plot
        plot_path = self.result_path / 'label_overlap.html'
        plot_path = PlotlyUtil.write_image_to_file(fig, plot_path)

        return ReportResult(
            name=self.name,
            info=f"Shows overlap between {self.row_label} and {self.column_label} labels.",
            output_figures=[ReportOutput(plot_path, "Label overlap heatmap")],
            output_tables=[ReportOutput(csv_path, "Label overlap counts")]
        )



import numpy as np


class EntropyCalculator:
    """
    Set of functions to help with diversity calculations
    """
    @staticmethod
    def shannon_entropy(x):
        a = np.array(x) if not isinstance(x, np.ndarray) else x
        a = a[np.nonzero(a)]
        p = a / np.sum(a)
        return -np.sum(p * np.log(p))

    @staticmethod
    def min_entropy(x):
        a = np.array(x) if not isinstance(x, np.ndarray) else x
        a = a[np.nonzero(a)]
        f = a / np.sum(a)
        return -np.log(np.max(f))

    @staticmethod
    def renyi_entropy(x, alpha):
        a = np.array(x) if not isinstance(x, np.ndarray) else x
        a = a[np.nonzero(a)]
        f = a / np.sum(a)
        if np.abs(alpha - 1) < 1e-10:
            entropy = EntropyCalculator.shannon_entropy(a)
        elif alpha == np.inf:
            entropy = EntropyCalculator.min_entropy(a)
        else:
            entropy = (1 / (1 - alpha)) * np.log(np.sum(f ** alpha))
        return entropy



import os
import shutil
from unittest import TestCase

from immuneML.caching.CacheType import CacheType
from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.reports.data_reports.RandomDataPlot import RandomDataPlot
from immuneML.util.PathBuilder import PathBuilder


class TestRandomDataPlot(TestCase):

    def setUp(self) -> None:
        os.environ[Constants.CACHE_TYPE] = CacheType.TEST.name

    def test_random_data_plot(self):
        path = PathBuilder.remove_old_and_build(EnvironmentSettings.tmp_test_path / "random_data_plot")

        params = {"n_points_to_plot": 10}
        report = RandomDataPlot.build_object(**params)

        # make sure to set the 'path' manually
        report.result_path = path

        self.assertTrue(report.check_prerequisites())
        result = report._generate()

        # ensure result files are generated
        self.assertTrue(os.path.isfile(result.output_figures[0].path))
        self.assertTrue(os.path.isfile(result.output_tables[0].path))

        # don't forget to remove the temporary path
        shutil.rmtree(path)



import argparse
import sys
from pathlib import Path

import yaml

from immuneML.util.PathBuilder import PathBuilder


def get_dataset_simulation_specs(args):
    dataset_specs = {"format": f"Random{args.dataset_type.title()}Dataset",
                     "params": {"labels": {f"{args.label_name}": {f"{args.class1_name}": int(args.class_balance) / 100,
                                                                  f"{args.class2_name}": (100 - int(
                                                                      args.class_balance)) / 100}},
                                f"{args.dataset_type}_count": int(args.count)}}

    return {args.dataset_name: dataset_specs}


def build_specs(args):
    specs = {
        "definitions": {
            "datasets": get_dataset_simulation_specs(args),
        },
        "instructions": {
            "my_dataset_generation_instruction": {
                "type": "DatasetExport",
                "datasets": [args.dataset_name],
            }
        }
    }

    return specs


def parse_commandline_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for building immuneML defintion YAML for Data Simulation tool")

    parser.add_argument("-d", "--dataset_name", default="dataset", help="The name of the dataset to export.")
    parser.add_argument("-t", "--dataset_type", choices=["receptor", "sequence", "repertoire"], required=True,
                        help="The type of the created dataset (receptor/sequence/repertoire).")
    parser.add_argument("-c", "--count", type=int, required=True, help="The number of examples to generate.")
    parser.add_argument("-l", "--label_name", required=True, help="The name of the label to assign to the dataset.")
    parser.add_argument("-x", "--class1_name", required=True,
                        help="The name of the first class to assign to the dataset.")
    parser.add_argument("-y", "--class2_name", required=True,
                        help="The name of the second class to assign to the dataset.")
    parser.add_argument("-b", "--class_balance", type=int, required=True,
                        help="The class balance percentage (in the range 0 - 100).")

    parser.add_argument("-o", "--output_path", required=True,
                        help="Output location for the generated yaml file (directiory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")

    return parser.parse_args(args)


def main(args):
    parsed_args = parse_commandline_arguments(args)
    specs = build_specs(parsed_args)

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    with output_location.open("w") as file:
        yaml.dump(specs, file)

    return str(output_location)


if __name__ == "__main__":
    main(sys.argv[1:])



from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Set, Dict

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.example_weighting.ExampleWeightingStrategy import ExampleWeightingStrategy
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.hyperparameter_optimization.states.HPAssessmentState import HPAssessmentState
from immuneML.hyperparameter_optimization.states.HPItem import HPItem
from immuneML.hyperparameter_optimization.strategy.HPOptimizationStrategy import HPOptimizationStrategy
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.reports.ReportResult import ReportResult


@dataclass
class TrainMLModelState:
    dataset: Dataset
    hp_strategy: HPOptimizationStrategy
    hp_settings: List[HPSetting]
    assessment: SplitConfig
    selection: SplitConfig
    metrics: Set[ClassificationMetric]
    optimization_metric: ClassificationMetric
    label_configuration: LabelConfiguration
    path: Path = None
    context: dict = None
    number_of_processes: int = 1
    reports: dict = field(default_factory=dict)
    name: str = None
    refit_optimal_model: bool = None
    export_all_ml_settings: bool = None
    example_weighting: ExampleWeightingStrategy = None
    sequence_type: SequenceType = SequenceType.AMINO_ACID
    region_type: RegionType = RegionType.IMGT_CDR3
    optimal_hp_items: Dict[str, HPItem] = field(default_factory=dict)
    optimal_hp_item_paths: Dict[str, str] = field(default_factory=dict)
    assessment_states: List[HPAssessmentState] = field(default_factory=list)
    report_results: List[ReportResult] = field(default_factory=list)



import argparse
from pathlib import Path

from immuneML.data_model.bnp_util import read_yaml, write_yaml


def make_yamls(original_yaml_path: Path, result_yaml_path: Path):
    """
    Maps the section with MatchedSequencesEncoder in the yaml to multiple yaml files

    From this:
        definitions:
            encodings:
                my_ms_encoding:
                    MatchedSequences:
                        reference:
                            format: VDJDB
                            params:
                                path: [path/to/file.csv, path2.csv, path3.csv]
    To:

        file 1:

        definitions:
            encodings:
                my_ms_encoding:
                    MatchedSequences:
                        reference:
                            format: VDJDB
                            params:
                                path: path/to/file.csv

        another file:

        definitions:
            encodings:
                my_ms_encoding:
                    MatchedSequences:
                        reference:
                            format: VDJDB
                            params:
                                path: path2.csv


    Args:
        original_yaml_path: Path to the original file
        result_yaml_path: Path where the resulting yaml files will be stored

    Returns: None

    """
    original_yaml = read_yaml(original_yaml_path)

    assert "definitions" in original_yaml and "encodings" in original_yaml[
        'definitions'], f'Format not supported: {original_yaml}'
    assert all([list(enc.keys())[0] == "MatchedSequences" for enc in
                original_yaml['definitions']['encodings'].values()]), f"Encoding not supported: {original_yaml}"

    encoding_name = list(original_yaml['definitions']['encodings'].keys())[0]

    for i, path in enumerate(
            original_yaml['definitions']['encodings'][encoding_name]["MatchedSequences"]['reference']['params'][
                'path']):
        print(path)
        tmp_yaml = original_yaml.copy()
        tmp_yaml['definitions']['encodings'][encoding_name]["MatchedSequences"]['reference']['params']['path'] = path
        write_yaml(result_yaml_path / f'ref{i}.yaml', tmp_yaml)
        print(f"Created {result_yaml_path / f'ref{i}.yaml'}")


def main():
    parser = argparse.ArgumentParser(description="immuneML yaml helper tool to expand ref sequence sets")
    parser.add_argument("original_path", help="Path to specification YAML file. Always used to define the "
                                              "analysis.")
    parser.add_argument("result_path", help="Output directory path.")

    namespace = parser.parse_args()
    make_yamls(Path(namespace.original_path), Path(namespace.result_path))


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from pathlib import Path

@dataclass
class CompAIRRParams:
    compairr_path: Path
    keep_compairr_input: bool
    differences: int
    indels: bool
    ignore_counts: bool
    ignore_genes: bool
    threads: int
    output_filename: str
    log_filename: str
    output_pairs: bool
    pairs_filename: str
    is_cdr3: bool = False
    do_repertoire_overlap: bool = True
    do_sequence_matching: bool = False



from pathlib import Path

from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.workflows.instructions.ligo_simulation.LigoSimInstruction import LigoSimInstruction


class LigoSimParser:

    def parse(self, key: str, instruction: dict, symbol_table: SymbolTable, path: Path = None) -> LigoSimInstruction:

        location = LigoSimParser.__name__
        keys = ["simulation", "type", 'sequence_batch_size', "max_iterations", "export_p_gens",
                "number_of_processes"]
        ParameterValidator.assert_keys(instruction.keys(), keys, location, key)

        for param_key in ['export_p_gens']:
            ParameterValidator.assert_type_and_value(instruction[param_key], bool, location, param_key)
        for param_key in ['max_iterations', 'sequence_batch_size', 'number_of_processes']:
            ParameterValidator.assert_type_and_value(instruction[param_key], int, location, param_key, 1)

        simulation = get_simulation_from_symbol_table(instruction['simulation'], symbol_table, location)

        params = {**{key: value for key, value in instruction.items() if key != 'type'},
                  **{'simulation': simulation, 'signals': symbol_table.get_signals(), 'name': key}}
        instruction = LigoSimInstruction(**params)
        return instruction


def get_simulation_from_symbol_table(sim_key, symbol_table, location):
    ParameterValidator.assert_type_and_value(sim_key, str, location, 'simulation')
    ParameterValidator.assert_in_valid_list(sim_key,
                                            [sim.item.identifier for sim in
                                             symbol_table.get_by_type(SymbolType.SIMULATION)],
                                            location, 'simulation')
    return symbol_table.get(sim_key)


import logging
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import plotly.express as px

from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.dsl.definition_parsers.MLParser import MLParser
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class DimensionalityReduction(EncodingReport):
    """
    This report visualizes the data obtained by dimensionality reduction. The data points can be highlighted by label of
    interest. It is also possible to specify labels that contain lists of values (e.g., HLA), in which case the data points
    will be duplicated (so that each point refers to one HLA allele) and jittered slightly to improve visibility
    before being highlighted by the concrete HLA allele values.

    **Specification arguments:**

    - labels (list): names of the label to use for highlighting data points; or None

    - dim_red_method (str): dimensionality reduction method to be used for plotting; if set, in a workflow, this
      dimensionality reduction will be used for plotting instead of any other set in the workflow; if None, it will
      visualize the encoded data of reduced dimensionality if set


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                rep1:
                    DimensionalityReduction:
                        labels: [epitope, source]
                        dim_red_method:
                            PCA:
                                n_components: 2

    """

    @classmethod
    def build_object(cls, **kwargs):
        if "dim_red_method" in kwargs and kwargs['dim_red_method'] and kwargs['dim_red_method'] != 'None':
            cls_name = list(kwargs['dim_red_method'].keys())[0]
            method = MLParser.parse_any_model("dim_red_method", kwargs['dim_red_method'], cls_name)[0]
        else:
            method = None

        location = f"DimensionalityReduction ({kwargs['name'] if 'name' in kwargs else ''})"

        # backwards compatibility: to be removed from next major version
        if "label" in kwargs:
            ParameterValidator.warn_deprecated_parameter("label", "labels", location)
            ParameterValidator.assert_type_and_value(kwargs["label"], str, location, "label")
            labels = [kwargs["label"]]
            del kwargs["label"]
        else:
            ParameterValidator.assert_type_and_value(kwargs["labels"], list, location, "labels")
            labels = kwargs["labels"]
            ParameterValidator.assert_all_type_and_value(labels, str, location, "labels")

        return DimensionalityReduction(**{**kwargs, "dim_red_method": method, 'labels': labels})

    def __init__(self, dataset: Dataset = None, batch_size: int = 1, result_path: Path = None,
                 name: str = None, labels: list = None, dim_red_method: DimRedMethod = None):
        super().__init__(dataset=dataset, result_path=result_path, name=name)
        self._labels = labels
        self._dim_red_method = dim_red_method
        self._dimension_names = ['dimension_1', 'dimension_2'] if self._dim_red_method is None \
            else self._dim_red_method.get_dimension_names()
        self.info = ("This report visualizes the encoded data after applying dimensionality reduction dim_red,"
                     " optionally colored by labels of interest.")

    def check_prerequisites(self):
        return (isinstance(self.dataset.encoded_data, EncodedData) and
                (self.dataset.encoded_data.dimensionality_reduced_data is not None or self._dim_red_method is not None))

    def _generate(self) -> ReportResult:
        if self._dim_red_method:
            assert self.dataset.encoded_data.examples is not None, \
                f"{DimensionalityReduction.__name__}: data not encoded, report will not be made."
            dim_reduced_data = self._dim_red_method.fit_transform(self.dataset)
        else:
            assert self.dataset.encoded_data.dimensionality_reduced_data is not None
            dim_reduced_data = self.dataset.encoded_data.dimensionality_reduced_data

        assert dim_reduced_data.shape[1] == 2, \
            (f"{DimensionalityReduction.__name__}: {self.name}: dimensionality reduced data is not 2d (got: "
             f"{dim_reduced_data.shape}, so it cannot be plotted.")
        data_labels = None

        try:
            data_labels = self.dataset.get_metadata(self._labels, return_df=True)[self._labels]
        except (AttributeError, TypeError) as e:
            logging.warning(f"Labels {self._labels} not found in the dataset. Skipping label coloring in the plot.")

        PathBuilder.build(self.result_path)

        df = pd.DataFrame({'example_id': self.dataset.get_example_ids(),
                           self._dimension_names[0]: dim_reduced_data[:, 0],
                           self._dimension_names[1]: dim_reduced_data[:, 1]})

        if hasattr(self.dataset, 'get_metadata_fields') and 'subject_id' in self.dataset.get_metadata_fields():
            df['subject_id'] = self.dataset.get_metadata(['subject_id'], return_df=True)['subject_id']

        if self._labels:
            df[self._labels] = data_labels
        df.to_csv(self.result_path / 'dimensionality_reduced_data.csv', index=False)

        report_output_figures = self._safe_plot(df=df, output_written=True)

        dim_red_text = f" ({self._dim_red_method.__class__.__name__})" if self._dim_red_method else ""

        return ReportResult(name=self.name, info=self.info.replace(" dim_red", dim_red_text),
                            output_figures=report_output_figures,
                            output_tables=[ReportOutput(self.result_path / 'dimensionality_reduced_data.csv',
                                                        'data after dimensionality reduction')])

    def _plot(self, df: pd.DataFrame) -> List[ReportOutput]:
        PathBuilder.build(self.result_path)
        outputs = []
        if self._labels:
            for label in self._labels:

                df_copy = self._parse_labels_with_lists(df, label)

                unique_values = df_copy[label].unique()

                hover_data = self._dimension_names + self._labels
                if 'subject_id' in df_copy.columns:
                    hover_data += ['subject_id']
                elif 'example_id' in df_copy.columns:
                    hover_data += ['example_id']

                if len(unique_values) <= 15:
                    df_copy[label] = df_copy[label].astype('category')
                    figure = px.scatter(df_copy, x=self._dimension_names[0], y=self._dimension_names[1], color=label,
                                        color_discrete_sequence=px.colors.qualitative.Vivid,
                                        hover_data=hover_data,
                                        category_orders={label: sorted(unique_values)})
                else:
                    figure = px.scatter(df_copy, x=self._dimension_names[0], y=self._dimension_names[1], color=label,
                                        hover_data=hover_data)

                figure.update_layout(template="plotly_white", showlegend=True)
                figure.update_traces(opacity=.6)

                file_path = self.result_path / f"dimensionality_reduction_{label}.html"
                file_path = PlotlyUtil.write_image_to_file(figure, file_path)
                outputs.append(ReportOutput(path=file_path,
                                            name="Data visualization after dimensionality reduction "
                                                 "(highlighted by {})".format(label)))
        else:
            # No label case - just plot points
            figure = px.scatter(df, x=self._dimension_names[0], y=self._dimension_names[1])
            figure.update_layout(template="plotly_white")
            figure.update_traces(opacity=.6)

            file_path = self.result_path / "dimensionality_reduction.html"
            file_path = PlotlyUtil.write_image_to_file(figure, file_path)
            outputs.append(ReportOutput(path=file_path, name="Data visualization after dimensionality reduction"))

        return outputs

    def _parse_labels_with_lists(self, df: pd.DataFrame, label: str) -> pd.DataFrame:
        df_long = df.copy()

        df_long[label] = df_long[label].apply(parse_list_column)

        if isinstance(df_long[label].iloc[0], (list, tuple)):
            df_long = df_long.explode(label)

            # Compute jitter based on the range of each axis
            x_range = df_long[self._dimension_names[0]].max() - df_long[self._dimension_names[0]].min()
            y_range = df_long[self._dimension_names[1]].max() - df_long[self._dimension_names[1]].min()
            jitter_strength = 0.005 * min(x_range, y_range)

            # Apply jitter
            df_long[self._dimension_names[0]] += np.random.uniform(-jitter_strength, jitter_strength, size=len(df_long))
            df_long[self._dimension_names[1]] += np.random.uniform(-jitter_strength, jitter_strength, size=len(df_long))

        return df_long


def parse_list_column(value):
    """Parses a string representation of a list or tuple into an actual list."""
    if not value or pd.isna(value):
        return 'unknown'
    if isinstance(value, str):
        value = value.strip()
        if (value.startswith('"') and value.endswith('"')) or (value.startswith('\'') and value.endswith('\'')):
            value = value[1:-1]
            items = [item.strip().replace('\'', '') for item in value.split(',') if item.strip()]
            return items
    return value



import re


class StringHelper:

    @staticmethod
    def camel_case_to_words(camel_case_string: str):
        string = camel_case_string[0].upper() + camel_case_string[1:]
        return re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', string)

    @staticmethod
    def camel_case_to_word_string(camel_case_string: str):
        return " ".join(StringHelper.camel_case_to_words(camel_case_string))

    @staticmethod
    def pad_sequence_in_the_middle(sequence: str, max_len: int, pad_char: str) -> str:
        if len(sequence) == max_len:
            return sequence
        else:
            pad_start = len(sequence) // 2
            pad_len = max_len - len(sequence)
            return sequence[:pad_start] + pad_char * pad_len + sequence[pad_start:]



from typing import List

import pandas as pd
import bionumpy as bnp

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.bnp_util import get_sequence_field_name
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.BackgroundSequences import BackgroundSequences
from immuneML.simulation.implants.Signal import Signal
from immuneML.simulation.util.util import annotate_sequences, make_annotated_dataclass


def annotate_sequence_dataset(dataset: SequenceDataset, signals: List[Signal],
                              region_type: RegionType = RegionType.IMGT_CDR3,
                              sequence_type: SequenceType = SequenceType.AMINO_ACID) -> pd.DataFrame:
    data = dataset.data
    sequences = BackgroundSequences(sequence_aa=getattr(data, get_sequence_field_name(region_type,
                                                                            SequenceType.AMINO_ACID)),
                                    sequence=getattr(data,
                                                     get_sequence_field_name(region_type, SequenceType.NUCLEOTIDE)),
                                    v_call=data.v_call, j_call=data.j_call,
                                    region_type=[region_type.value for _ in range(len(data))],
                                    frame_type=['' for _ in range(len(data))], p_gen=[-1 for _ in range(len(data))],
                                    duplicate_count=[1 for _ in range(len(data))], locus=data.locus,
                                    from_default_model=[0 for _ in range(len(data))])

    annotated_dataclass = make_annotated_dataclass([(signal.id, int) for signal in signals] +
                                                   [(f'{signal.id}_positions', str) for signal in signals] +
                                                   [('signals_aggregated', str)], signals)

    seqs_with_signal_matrix = annotate_sequences(sequences=sequences,
                                                 is_amino_acid=sequence_type == SequenceType.AMINO_ACID,
                                                 all_signals=signals,
                                                 annotated_dc=annotated_dataclass,
                                                 region_type=region_type)

    annotated_sequences = seqs_with_signal_matrix.topandas()
    annotated_sequences.drop(columns=['p_gen', 'frame_type', 'region_type', 'duplicate_count', 'from_default_model'],
                             inplace=True)
    annotated_sequences.rename(columns={'sequence_aa': get_sequence_field_name(region_type, SequenceType.AMINO_ACID),
                                        'sequence': get_sequence_field_name(region_type, SequenceType.NUCLEOTIDE)},
                               inplace=True)

    if hasattr(data, 'data_origin'):
        annotated_sequences['data_origin'] = data.data_origin.tolist()

    return annotated_sequences



import re
import shutil
from pathlib import Path

import numpy as np
import pandas as pd


def make_skewed_model_files(v_genes: list, j_genes: list, original_model_path: Path, new_model_path: Path) -> Path:
    """

    Args:
        j_genes: a list of j gene patterns to be matched against the original model
        v_genes: a list of v gene patterns to be matched against the original model
        original_model_path: path to the folder of the original igor model files
        new_model_path: where to store updated files

    Returns:
        Path to the folder with new model files

    """

    genes = _import_genes_from_model_params(original_model_path)
    _check_if_all_genes_present(genes, v_genes, j_genes)

    if len(v_genes) != 0:
        v_gene_mask = np.array([any(v_gene in gene for v_gene in v_genes) for gene in genes['v']])
        v_gene_weights = v_gene_mask / v_gene_mask.sum()
    else:
        v_gene_weights = None

    if len(j_genes):
        j_gene_mask = np.array([any(j_gene in gene for j_gene in j_genes) for gene in genes['j']])
        j_gene_weights = j_gene_mask / j_gene_mask.sum()
    else:
        j_gene_weights = None

    _make_new_marginals_file(original_model_path, new_model_path, v_gene_weights, j_gene_weights)
    _copy_remaining_model_files(original_model_path, new_model_path)

    return new_model_path


def _copy_remaining_model_files(original_model_path: Path, new_model_path: Path):
    for filename in ['J_gene_CDR3_anchors.csv', 'V_gene_CDR3_anchors.csv', 'model_params.txt']:
        shutil.copy(original_model_path / filename, new_model_path / filename)


def _make_new_marginals_file(original_model_path: Path, new_model_path, v_gene_weights, j_gene_weights):
    with (original_model_path / 'model_marginals.txt').open('r') as file:
        marginals_text = file.readlines()

    if v_gene_weights is not None:
        marginals_text[3] = _get_new_independent_genes(marginals_text, v_gene_weights, gene_name='v', index=0)
    if j_gene_weights is not None:
        marginals_text = _parse_j_genes(marginals_text, j_gene_weights)

    with (new_model_path / 'model_marginals.txt').open('w') as file:
        file.writelines(marginals_text)


def _parse_j_genes(marginals_text: list, j_gene_weights):
    dim_pattern = re.compile("\$Dim\[(\d+)(\,)*(\d*)\]")

    assert "@j_choice" in marginals_text[4]
    hit = re.match(dim_pattern, marginals_text[5])
    assert hit
    if hit.group(3) and hit.group(3) != "":
        marginals_text[5:] = _update_conditional_j_genes(marginals_text[5:], j_gene_weights)
    else:
        marginals_text[7] = _get_new_independent_genes(marginals_text, j_gene_weights, gene_name='j', index=4)

    return marginals_text


def _update_conditional_j_genes(marginals_trunc_list, j_gene_weights) -> list:
    index = 0

    while "@" not in marginals_trunc_list[index] and index < len(marginals_trunc_list):
        assert f"#[v_choice,{index // 2}]" in marginals_trunc_list[index]
        marginals_trunc_list[index + 1] = _recompute_gene_proba_from_line(marginals_trunc_list[index + 1], j_gene_weights)
        index += 2

    return marginals_trunc_list


def _get_new_independent_genes(marginals_text: list, gene_weights, gene_name: str, index: int):
    assert f"@{gene_name}_choice" in marginals_text[index]
    return _recompute_gene_proba_from_line(marginals_text[index + 3], gene_weights)


def _recompute_gene_proba_from_line(line, gene_weights) -> str:
    original_probabilities = np.array(line.split("%")[1].split(","), dtype=float)
    updated_probabilities = original_probabilities * gene_weights
    updated_probabilities = (updated_probabilities / updated_probabilities.sum()).astype(str)
    return "%" + ",".join(updated_probabilities) + "\n"


def _check_if_all_genes_present(original_model_genes, signal_v_genes: list, signal_j_genes: list):
    for gene_name, genes in {'v': signal_v_genes, 'j': signal_j_genes}.items():
        assert all(any(gene in org_gene for org_gene in original_model_genes[gene_name]) for gene in genes), \
            f"Not all {gene_name.upper()} genes in signals are present in the original generative model."


def _import_genes_from_anchor_files(original_model_path: Path) -> dict:
    genes = {"v": [], "j": []}

    for key in genes.keys():
        df = pd.read_csv(original_model_path / f"{key.upper()}_gene_CDR3_anchors.csv")
        genes[key] = df['gene'].values

    return genes


def _import_genes_from_model_params(original_model_path: Path) -> dict:
    genes = {"v": [], "j": []}

    with (original_model_path / 'model_params.txt').open('r') as file:
        lines = file.readlines()

    index = 0

    for gene_name in ["v", "j"]:
        while f"{gene_name.upper()}_gene;" not in lines[index] and lines[index] != "":
            index += 1
        index += 1
        line = lines[index]
        while line[0] != "#" and line != "":
            gene = line.split("%")[1].split(";")[0]
            gene_index = int(line.split("%")[1].split(";")[-1].replace("\n", ""))
            assert gene_name.upper() in gene, (gene, gene_name + " gene")
            genes[gene_name].append((gene, gene_index))
            index += 1
            line = lines[index]

        genes[gene_name] = sorted(genes[gene_name], key=lambda g: g[1])
        genes[gene_name] = [g[0] for g in genes[gene_name]]

    return genes



import logging
from multiprocessing.pool import Pool
from pathlib import Path

import dill
import numpy as np

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.preprocessing.filters.Filter import Filter


class CountPerSequenceFilter(Filter):
    """
    Removes all sequences from a Repertoire when they have a count below low_count_limit, or sequences with no count
    value if remove_without_counts is True. This filter can be applied to Repertoires and RepertoireDatasets.

    **Specification arguments:**

    - low_count_limit (int): The inclusive minimal count value in order to retain a given sequence.

    - remove_without_count (bool): Whether the sequences without a reported count value should be removed.

    - remove_empty_repertoires (bool): Whether repertoires without sequences should be removed.
      Only has an effect when remove_without_count is also set to True. If this is true, this preprocessing cannot be used with :ref:`TrainMLModel`
      instruction, but only with :ref:`DatasetExport` instruction instead.

    - batch_size (int): number of repertoires that can be loaded at the same time (only affects the speed when applying this filter on a RepertoireDataset)


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    CountPerSequenceFilter:
                        remove_without_count: True
                        remove_empty_repertoires: True
                        low_count_limit: 3
                        batch_size: 4

    """

    def __init__(self, low_count_limit: int, remove_without_count: bool, remove_empty_repertoires: bool, batch_size: int, result_path: Path = None):
        super().__init__(result_path)
        self.low_count_limit = low_count_limit
        self.remove_without_count = remove_without_count
        self.remove_empty_repertoires = remove_empty_repertoires
        self.batch_size = batch_size

    def keeps_example_count(self) -> bool:
        return not self.remove_empty_repertoires

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1) -> RepertoireDataset:
        self.check_dataset_type(dataset, [RepertoireDataset], "CountPerSequenceFilter")
        self.result_path = result_path if result_path is not None else self.result_path

        with Pool(self.batch_size) as pool:
            repertoires = pool.map(self._process_repertoire, [dill.dumps(rep) for rep in dataset.repertoires])

        if self.remove_empty_repertoires:
            repertoires = self._remove_empty_repertoires(repertoires)

        processed_dataset = RepertoireDataset.build_from_objects(repertoires=repertoires, path=result_path)

        self.check_dataset_not_empty(processed_dataset, "CountPerSequenceFilter")

        return processed_dataset

    def _process_repertoire(self, repertoire: Repertoire) -> Repertoire:

        if isinstance(repertoire, bytes):
            repertoire = dill.loads(repertoire)

        counts = repertoire.data.duplicate_count
        not_none_indices = counts != -1
        counts[not_none_indices] = counts[not_none_indices].astype(int)
        indices_to_keep = np.full(repertoire.get_element_count(), False)
        if self.remove_without_count and self.low_count_limit is not None:
            np.greater_equal(counts, self.low_count_limit, out=indices_to_keep, where=not_none_indices)
        elif self.remove_without_count:
            indices_to_keep = not_none_indices
        elif self.low_count_limit is not None:
            indices_to_keep[np.logical_not(not_none_indices)] = True
            np.greater_equal(counts, self.low_count_limit, out=indices_to_keep, where=not_none_indices)

        processed_repertoire = Repertoire.build_like(repertoire, indices_to_keep, self.result_path, filename_base=f"{repertoire.data_filename.stem}_filtered")

        logging.info(f"{CountPerSequenceFilter.__name__}: finished processing repertoire "
                     f"(subject_id: {repertoire.metadata['subject_id'] if repertoire.metadata and 'subject_id' in repertoire.metadata else ''}, "
                     f"id: {repertoire.identifier}).")

        return processed_repertoire



import argparse
import sys

# Ensure the immuneML/ project 'root dir' is added to sys.path
# Adding "." and "../" allows the script to be run from immuneML/ and immuneML/scripts/
# When encountering ModuleNotFoundError, try adding the absolute path to the project 'root dir' here
sys.path.extend([".", "../"])

from scripts.checker_util import *
from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.dataset.SequenceDataset import SequenceDataset
from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.reports.encoding_reports.DesignMatrixExporter import DesignMatrixExporter
from immuneML.simulation.dataset_generation.RandomDatasetGenerator import RandomDatasetGenerator
from immuneML.util.ReflectionHandler import ReflectionHandler


def parse_commandline_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for testing new immuneML DatasetEncoder classes")

    usage_args = parser.add_argument_group('usage arguments')
    usage_args.add_argument("-e", "--encoder_file", type=str, required=True, help="Path to the (dataset-specific) encoder file, placed in the correct immuneML subfolder. ")
    usage_args.add_argument("-d", "--dataset_type", type=str, choices=["repertoire", "sequence", "receptor"], required=True, help="Whether to test using 'sequence', 'receptor' or 'repertoire' dataset.")
    usage_args.add_argument("-p", "--no_default_parameters",  action='store_true', help="If enabled, it is assumed that no default parameters file exists, and the Encoder can be run without supplying additional parameters. ")

    logging_args = parser.add_argument_group('logging arguments')
    logging_args.add_argument("-l", "--log_file", type=str, default="check_new_encoder_log.txt", help="Path to the output log file. If already present, the file will be overwritten (default='./check_new_encoder_log.txt').")
    logging_args.add_argument("-t", "--tmp_path", type=str, default="./tmp", help="Path to the temporary output folder. If already present, the folder will be overwritten (default='./tmp').")

    return parser.parse_args(args)

def check_methods(encoder_instance):
    logging.info("Checking methods...")
    mssg = "Error: class method '{}' should not be overwritten from DatasetEncoder. Please remove this method."

    assert DatasetEncoder.store == encoder_instance.__class__.store, mssg.format("store")
    assert DatasetEncoder.store_encoder == encoder_instance.__class__.store_encoder, mssg.format("store_encoder")
    assert DatasetEncoder.load_attribute == encoder_instance.__class__.load_attribute, mssg.format("load_attribute")

    check_base_vs_instance_methods(DatasetEncoder, encoder_instance)

    logging.info("...Checking methods done.")

def check_encoded_dataset(encoded_dataset_result, dummy_dataset, encoder_instance):
    assert isinstance(encoded_dataset_result, dummy_dataset.__class__), f"Error: expected {encoder_instance.__class__.__name__}.encode() method to return an instance of {dummy_dataset.__class__.__name__} (with attached EncodedData object), found {encoded_dataset_result.__class__.__name__}"
    assert dummy_dataset.get_example_ids() == encoded_dataset_result.get_example_ids(), f"Error: expected the encoded dataset to be a copy of the original dataset (with attached EncodedData object). Found a difference in the example ids: {dummy_dataset.get_example_ids()} != {encoded_dataset_result.get_example_ids()}"

def check_encoded_data(encoded_data, dummy_dataset, base_class_name):
    assert isinstance(encoded_data, EncodedData), f"Error: expected the .encoded_data field of the output dataset to be an EncodedData object, found {encoded_data.__class__.__name__}"

    assert encoded_data.examples is not None, f"Error: EncodedData.examples is None, but should be a numeric matrix with a number of rows equal to the number of examples in the dataset ({dummy_dataset.get_example_count()})"
    assert encoded_data.examples.shape[0] == dummy_dataset.get_example_count(), f"Error: the number of rows in EncodedData.examples ({encoded_data.examples.shape[0]}) must be equal to the number of examples in the dataset ({dummy_dataset.get_example_count()})"

    assert encoded_data.example_ids == dummy_dataset.get_example_ids(), f"Error: EncodedData.example_ids must match the original dataset: {dummy_dataset.get_example_ids()}, found {encoded_data.example_ids}"
    assert encoded_data.encoding == base_class_name, f"Error: EncodedData.encoding must be set to the base class name ('{base_class_name}'), found {encoded_data.encoding}"

    if encoded_data.feature_names is None:
        logging.warning("EncodedData.feature_names is set to None. Please consider adding meaningful feature names to your encoding, as some Reports may crash without feature names.")

def get_dummy_dataset_labels(dummy_dataset):
    if isinstance(dummy_dataset, RepertoireDataset) or isinstance(dummy_dataset, ReceptorDataset):
        return [element.metadata["my_label"] for element in dummy_dataset.get_data()]
    elif isinstance(dummy_dataset, SequenceDataset):
        return [element.get_attribute("my_label") for element in dummy_dataset.get_data()]

def check_labels(encoded_data, dummy_dataset):
    labels_descr = "\nLabels must be a dictionary, where: " \
                   "\n- the keys are the label names as found by EncoderParams.label_config.get_labels_by_name()" \
                   "\n- the values are the label values for each example, which must follow the same order as examples in the dataset. Label values can be retrieved using: Repertoire.metadata[label_name], Receptor.metadata[label_name], or Sequence.get_attribute(label_name)."
    assert encoded_data.labels is not None, f"Error: EncodedData.labels is None, but labels must be set. " + labels_descr
    assert isinstance(encoded_data.labels, dict), f"Error: EncodedData.labels is of type {type(encoded_data.labels)}, but must be a dictionary. " + labels_descr

    assert "unnecessary_label" not in encoded_data.labels.keys(), "Error: EncodedData.labels contains additional dataset keys which were not specified for encoding. Please make sure only to use the labels given by EncoderParams.label_config.get_labels_by_name()"

    expected = {"my_label": get_dummy_dataset_labels(dummy_dataset)}
    assert set(encoded_data.labels.keys()) == {"my_label"}, f"Error: label keys are incorrect. Expected to find the label 'my_dataset', found {encoded_data.labels.keys()}. Please make sure only to use the labels given by EncoderParams.label_config.get_labels_by_name()"
    assert encoded_data.labels["my_label"] == expected["my_label"], f"Error: values for label 'my_label' are incorrect. Expected  {encoded_data.labels['my_label']}, found {expected['my_label']}. Please make sure the label values follow the same order as examples in the dataset. Label values can be retrieved using: Repertoire.metadata[label_name], Receptor.metadata[label_name], or Sequence.get_attribute(label_name)."

def check_encode_method_with_labels(encoder_instance, dummy_dataset, base_class_name, tmp_path):
    logging.info(f"Testing encode() method functionality with labels...")

    lc = LabelConfiguration()
    lc.add_label(label_name="my_label", values=["yes", "no"], positive_class="yes")
    params = EncoderParams(result_path=Path(tmp_path) / "encoder_results",
                           label_config=lc,
                           encode_labels=True)
    encoded_dataset_result = encoder_instance.encode(dummy_dataset, params)

    check_encoded_dataset(encoded_dataset_result, dummy_dataset, encoder_instance)
    check_encoded_data(encoded_dataset_result.encoded_data, dummy_dataset, base_class_name)
    check_labels(encoded_dataset_result.encoded_data, dummy_dataset)

    logging.info("...Testing encode() method functionality with labels done.")

    return encoded_dataset_result

def check_encode_method_without_labels(encoder_instance, dummy_dataset, base_class_name, tmp_path):
    logging.info(f"Testing encode() method functionality without labels...")

    params = EncoderParams(result_path=Path(tmp_path) / "encoder_results_no_labels",
                           label_config=LabelConfiguration(),
                           encode_labels=False)
    encoded_dataset_result = encoder_instance.encode(dummy_dataset, params)

    check_encoded_dataset(encoded_dataset_result, dummy_dataset, encoder_instance)
    check_encoded_data(encoded_dataset_result.encoded_data, dummy_dataset, base_class_name)

    assert encoded_dataset_result.encoded_data.labels is None, f"Error: EncodedData.labels must be set to 'None' if EncoderParams.encode_labels is False. Found the following instead: {encoded_dataset_result.encoded_data.labels}"
    logging.info("...Testing encode() method functionality without labels done.")

    return encoded_dataset_result

def check_encode_method(encoder_instance, dummy_dataset, base_class_name, tmp_path):
    logging.info(f"Testing encode() method functionality...")

    encode_method_result_no_labels = check_encode_method_without_labels(encoder_instance, dummy_dataset, base_class_name, tmp_path)
    encode_method_result = check_encode_method_with_labels(encoder_instance, dummy_dataset, base_class_name, tmp_path)

    logging.info("...Testing encode() method functionality done.")

    return encode_method_result

def test_with_design_matrix_exporter(encoded_dataset_result, tmp_path):
    logging.info("Testing exporting the encoded dataset to .csv format with DesignMatrixExporter report...")

    dme_report = DesignMatrixExporter(dataset=encoded_dataset_result, result_path=Path(tmp_path) / "dme_report", file_format="csv")
    dme_report._generate()

    logging.info("...Testing exporting the encoded dataset to .csv format with DesignMatrixExporter report done.")

def check_encoder_class(encoder_class, dummy_dataset, default_params, base_class_name, tmp_path):
    assert issubclass(encoder_class, DatasetEncoder), f"Error: '{encoder_class.__name__}' is not a subclass of DatasetEncoder"

    logging.info(f"Attempting to run {encoder_class.__name__}.build_object() for datasets of type {dummy_dataset.__class__.__name__} (with default parameters if given)...")
    encoder_instance = encoder_class.build_object(dummy_dataset, **default_params)
    assert isinstance(encoder_instance, DatasetEncoder), f"Expected method 'build_object()' to return an instance of DatasetEncoder, found {type(encoder_instance)}. " + \
                                                         "It looks like the return statement may be missing." if encoder_instance is None else ""
    logging.info(f"...Successfully loaded an instance of {encoder_instance.__class__.__name__}")

    check_methods(encoder_instance)
    encoded_dataset_result = check_encode_method(encoder_instance, dummy_dataset, base_class_name, tmp_path)
    test_with_design_matrix_exporter(encoded_dataset_result, tmp_path=tmp_path)


def check_encoder_file(encoder_file):
    logging.info(f"Checking file name and location...")
    assert encoder_file.parent.parent.stem == "encodings", f"Error: The new DatasetEncoder file(s) must be located inside a new package (folder) under 'immuneML/encodings'. Found the following folder where 'encodings' was expected: {encoder_file.parent.parent}"

    assert encoder_file.suffix == ".py", f"Error: '{encoder_file}' is not a Python file"
    assert encoder_file.stem.endswith("Encoder"), f"Error: Filename '{encoder_file}' does not end with 'Encoder'"
    assert encoder_file.stem[0].isupper(), f"Error: Filename '{encoder_file}' does not start with an upper case character. Please rename the encoder (both class and filename) to '{encoder_file.stem.title()}'"

    check_is_alphanum_name(encoder_file.stem)

    init_file_path = encoder_file.parent / "__init__.py"
    if not init_file_path.is_file():
        with open(init_file_path, 'w') as file:
            pass
        logging.info(f"Expected an empty init file to be found at {init_file_path}. This file was not present previously, but has now been created.")

    logging.info(f"...Checking file name and location done.")

def get_encoder_class(encoder_file_path):
    logging.info(f"Attempting to load Encoder class named '{encoder_file_path.stem}'...")
    try:
        encoder_class = ReflectionHandler.get_class_from_path(encoder_file_path, class_name=encoder_file_path.stem)
    except AttributeError as e:
        raise AttributeError(f"Encountered an AttributeError while trying to load the class named {encoder_file_path.stem} from file {encoder_file_path}. "
                             f"Please double check if the class and filename match. ").with_traceback(e.__traceback__)

    logging.info(f"...Loading Encoder class done.")

    return encoder_class

def get_default_params_filepath_from_name(encoder_file_path):
    encoder_shortname = encoder_file_path.stem[:-7]
    return EnvironmentSettings.default_params_path / "encodings" / f"{DefaultParamsLoader.convert_to_snake_case(encoder_shortname)}_params.yaml"


def get_dummy_dataset(dataset_type, tmp_path):
    if dataset_type == "repertoire":
        return RandomDatasetGenerator.generate_repertoire_dataset(repertoire_count=5,
                                                                  sequence_count_probabilities={10: 1},
                                                                  sequence_length_probabilities={10: 1},
                                                                  labels={"my_label": {"yes": 0.5, "no": 0.5},
                                                                          "unnecessary_label": {"yes": 0.5, "no": 0.5}},
                                                                  path=Path(tmp_path))
    elif dataset_type == "receptor":
        return RandomDatasetGenerator.generate_receptor_dataset(receptor_count=5,
                                                                  chain_1_length_probabilities={10: 1},
                                                                  chain_2_length_probabilities={10: 1},
                                                                  labels={"my_label": {"yes": 0.5, "no": 0.5},
                                                                          "unnecessary_label": {"yes": 0.5, "no": 0.5}},
                                                                  path=Path(tmp_path))
    elif dataset_type == "sequence":
        return RandomDatasetGenerator.generate_sequence_dataset(sequence_count=5,
                                                                length_probabilities={10: 1},
                                                                labels={"my_label": {"yes": 0.5, "no": 0.5},
                                                                        "unnecessary_label": {"yes": 0.5, "no": 0.5}},
                                                                path=Path(tmp_path))


def run_checks(parsed_args):
    logging.info(f"Testing new encoder file: '{parsed_args.encoder_file}'")

    assert os.path.isfile(parsed_args.encoder_file), f"Error: '{parsed_args.encoder_file}' is not a file"
    encoder_file_path = Path(parsed_args.encoder_file)

    check_encoder_file(encoder_file_path)
    encoder_class = get_encoder_class(encoder_file_path)

    default_params_filepath = get_default_params_filepath_from_name(encoder_file_path)
    default_params = check_default_args(default_params_filepath, parsed_args.no_default_parameters, class_name=encoder_file_path.stem)
    dummy_dataset = get_dummy_dataset(parsed_args.dataset_type, parsed_args.tmp_path)
    check_encoder_class(encoder_class, dummy_dataset, default_params, base_class_name=encoder_file_path.stem, tmp_path=parsed_args.tmp_path)

    logging.info("\n\nCongratulations! If the script ran this far without encountering any error, "
                 f"\nyou have successfully integrated a new DatasetEncoder for {parsed_args.dataset_type} datasets into immuneML. "
                 "\n\nTo finalise your encoder: "
                 "\n- Consider testing the encoder with other dataset types, if applicable."
                 f"\n- Check the exported design matrix to make sure the encoded data fulfills your expectations. This can be found at: '{parsed_args.tmp_path}/dme_report/design_matrix.csv' "
                 "\n- Please take a look if any warnings were thrown in this log file, and consider resolving those. These issues are not critical for the performance of your DatasetEncoder, but may be necessary to resolve before merging the code into the main immuneML branch."
                 "\n- Add class documentation containing a general description of the encoding, its parameters, and an example YAML snippet. "
                 "\n- Add a unit test for the new encoder. "
                 "\n\nFor more details, see https://docs.immuneml.uio.no/latest/developer_docs/how_to_add_new_encoding.html")



def main(args):
    parsed_args = parse_commandline_arguments(args)

    setup_logger(parsed_args.log_file)
    set_tmp_path(parsed_args.tmp_path)

    try:
        run_checks(parsed_args)
    except Exception as e:
        logging.error("\n\nA critical error occurred when testing the new DatasetEncoder:\n")
        logging.exception(e)

        raise e


if __name__ == "__main__":
    main(sys.argv[1:])



from pathlib import Path
from typing import List

import pandas as pd
import plotly.graph_objects as go

from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_method_reports.ClusteringMethodReport import ClusteringMethodReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringItem


class ExternalLabelClusterSummary(ClusteringMethodReport):
    """
    This report summarizes the number of examples in a cluster with different values of external labels.
    For each external label, it creates:
    1. A contingency table showing the count of examples for each combination of cluster and label value
    2. A heatmap visualization of these counts

    It can be used in combination with Clustering instruction.

    **Specification arguments:**

    - external_labels (list): the list of metadata columns in the dataset that should be compared against cluster
      assignment

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        reports:
            my_external_label_cluster_summary:
                ExternalLabelClusterSummary:
                    external_labels: [disease, HLA]

    """

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys(list(kwargs.keys()), ['external_labels', 'name'],
                                       ExternalLabelClusterSummary.__name__, ExternalLabelClusterSummary.__name__)
        ParameterValidator.assert_all_type_and_value(kwargs['external_labels'], str,
                                                     ExternalLabelClusterSummary.__name__, 'external_labels')
        return ExternalLabelClusterSummary(**kwargs)

    def __init__(self, external_labels: List[str], name: str = None, item: ClusteringItem = None,
                 result_path: Path = None):
        super().__init__(name=name, result_path=result_path, clustering_item=item)
        self.external_labels = external_labels
        self.desc = "External Label Cluster Summary"

    def _generate(self) -> ReportResult:
        self.result_path = PathBuilder.build(self.result_path / self.name)
        report_outputs = self._process_analysis_results()

        if not report_outputs:
            return ReportResult(
                name=f"{self.desc} ({self.name})",
                info="No results were generated. This could be because no external labels were found in the dataset "
                     "metadata."
            )

        return ReportResult(
            name=f"{self.desc} ({self.name})",
            info="Summary of cluster assignments versus external labels",
            output_tables=[output for output in report_outputs if 'table' in output.name],
            output_figures=[output for output in report_outputs if 'heatmap' in output.name]
        )

    def _process_analysis_results(self) -> List[ReportOutput]:
        outputs = []

        predictions = self.item.predictions
        dataset = self.item.dataset

        # For each external label
        labels = dataset.get_metadata(self.external_labels, return_df=True)
        for label in self.external_labels:
            label_values = labels[label]

            # Create contingency table
            contingency_df = pd.crosstab(
                pd.Series(predictions, name='cluster'),
                pd.Series(label_values, name=label)
            )

            # Save contingency table
            table_path = self.result_path / f"{label}_contingency.csv"
            contingency_df.to_csv(table_path)
            outputs.append(ReportOutput(
                path=table_path,
                name=f"Contingency table for {label} ({self.item.cl_setting.get_key()})"
            ))

            # Create heatmap
            fig = go.Figure(data=go.Heatmap(
                z=contingency_df.values,
                x=contingency_df.columns,
                y=contingency_df.index,
                colorscale='Viridis',
                text=contingency_df.values,
                texttemplate='%{text}',
                hovertemplate='count: %{z}<br>cluster: %{y}<br>' + label + ': %{x}<extra></extra>',
                hoverongaps=False
            ))

            fig.update_layout(
                xaxis_title=label,
                yaxis_title='cluster',
                template='plotly_white'
            )

            fig.update_xaxes(type='category')
            fig.update_yaxes(type='category')

            heatmap_path = self.result_path / f"{label}_heatmap.html"
            plot_path = PlotlyUtil.write_image_to_file(fig, heatmap_path, contingency_df.shape[0])

            outputs.append(ReportOutput(
                path=plot_path,
                name=f"Distribution heatmap for {label} with example counts "
                     f"({dataset.get_example_count()} total examples)"
            ))

        return outputs

    def check_prerequisites(self):
        if self.item is None:
            return False

        if not self.external_labels:
            return False

        return True



import logging
from pathlib import Path
from typing import Callable, Tuple

import pandas as pd
from sklearn import metrics

from immuneML.ml_metrics import ClusteringMetric
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_reports.ClusteringReport import ClusteringReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringRunner import get_features, encode_dataset
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringState
from immuneML.workflows.instructions.clustering.ValidationHandler import get_complementary_classifier


class ClusteringStabilityReport(ClusteringReport):
    """
    Report that analyzes clustering stability by comparing results between discovery and validation datasets.
    The comparison uses a classifier-based approach where:
    1. A classifier is trained on discovery data using cluster assignments as labels
    2. Cluster assignments are predicted for validation data
    3. Predictions are compared with actual validation clustering results using the specified similarity metric

    This report can be used with the Clustering instruction under 'reports'.

    **Specification arguments:**

    - metric (str): Name of any clustering evaluation metric from sklearn.metrics that compares two sets of labels
      (e.g., adjusted_rand_score, jaccard_score, adjusted_mutual_info_score, normalized_mutual_info_score).
      If an invalid metric name is provided, defaults to adjusted_rand_score.


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_clustering_instruction:
            type: Clustering
            reports:
                my_stability_report:
                    ClusteringStabilityReport:
                        metric: jaccard_score

    """

    DEFAULT_METRIC = "adjusted_rand_score"

    @classmethod
    def build_object(cls, **kwargs):
        location = "ClusteringStabilityReport"
        name = kwargs["name"] if "name" in kwargs else None
        similarity_metric = kwargs.get("metric", cls.DEFAULT_METRIC)

        ParameterValidator.assert_type_and_value(similarity_metric, str, location, "metric")
        assert ClusteringMetric.is_valid_metric(similarity_metric), f"Invalid metric '{similarity_metric}' provided. " \
                                                                    f"Please use a valid clustering evaluation metric from sklearn.metrics."

        return ClusteringStabilityReport(name=name, similarity_metric=similarity_metric)

    def __init__(self, similarity_metric: str, name: str = None, state: ClusteringState = None,
                 result_path: Path = None, number_of_processes: int = 1):
        super().__init__(name=name, result_path=result_path, number_of_processes=number_of_processes)
        self.state = state
        self.metric_fn, self.metric = self._get_similarity_metric(similarity_metric)
        self.result_name = "clustering_stability_analysis"
        self.desc = "Clustering Stability Report"

    def _get_similarity_metric(self, metric_name: str) -> Tuple[Callable, str]:
        try:
            metric_func = getattr(metrics, metric_name)
            return metric_func, metric_name
        except AttributeError:
            logging.warning(f"Invalid similarity metric '{metric_name}', defaulting to {self.DEFAULT_METRIC}")
            return getattr(metrics, self.DEFAULT_METRIC), self.DEFAULT_METRIC

    def _generate(self) -> ReportResult:
        """Generate the stability analysis report"""
        self.result_path = PathBuilder.build(self.result_path / self.name)
        stability_results = []

        # For each data split
        for split_id in range(self.state.config.split_config.split_count):
            # For each clustering setting
            for setting in self.state.config.clustering_settings:
                stability_results.append(self.compute_stability_score_for_setting(setting, split_id))

        if not stability_results:
            logging.warning("No stability results could be calculated. Check if validation data is available.")
            return ReportResult(
                name=f"{self.desc} ({self.name})",
                info="No stability results could be calculated. Check if validation data is available."
            )
        else:
            df = pd.DataFrame(stability_results)

            table_path = self.result_path / f"{self.result_name}.csv"
            df.to_csv(table_path, index=False)
            table_output = ReportOutput(path=table_path, name="Stability scores per setting")

            return ReportResult(
                name=f"{self.desc} ({self.name})",
                info=f"The report analyzes clustering stability by comparing results between discovery and validation "
                     f"datasets. The comparison uses a classifier-based approach where "
                     f"a classifier is trained on discovery data using cluster assignments as labels, cluster "
                     f"assignments are predicted for validation data and then the predictions are compared with actual "
                     f"validation clustering results using {self.metric}.",
                output_figures=[self.make_figure(df)],
                output_tables=[table_output]
            )

    def make_figure(self, df: pd.DataFrame) -> ReportOutput:
        import plotly.express as px
        fig = px.box(df, x='clustering_setting', y=self.metric, points='all', color='clustering_setting',
                     color_discrete_sequence=px.colors.qualitative.Vivid)
        fig.update_layout(xaxis_title="clustering setting", showlegend=False,
                          yaxis_title=self.metric,
                          template="plotly_white")
        fig.update_traces(marker=dict(opacity=0.5), jitter=0.3)

        plot_path = PlotlyUtil.write_image_to_file(fig, self.result_path / f"stability_boxplot.html", df.shape[0])

        return ReportOutput(plot_path,
                            name=f"Clustering Stability Analysis ({self.metric})")

    def compute_stability_score_for_setting(self, setting, split_id: int):

        setting_key = setting.get_key()

        # Get discovery and validation results
        discovery_result = self._get_cluster_result('discovery', split_id, setting_key)
        validation_result = self._get_cluster_result('method_based_validation', split_id, setting_key)

        if discovery_result and validation_result:
            discovery_clusters = discovery_result.item.predictions
            validation_clusters = validation_result.item.predictions

            # Get encoded data
            discovery_data = get_features(discovery_result.item.dataset, setting)
            validation_data = get_features(
                encode_dataset(validation_result.item.dataset, setting, self.number_of_processes,
                               label_config=self.state.config.label_config, learn_model=False,
                               sequence_type=self.state.config.sequence_type,
                               region_type=self.state.config.region_type,
                               encoder=discovery_result.item.encoder), setting)

            if discovery_data is not None and validation_data is not None:
                try:
                    # Train classifier on discovery data
                    clf = get_complementary_classifier(setting)
                    clf.fit(discovery_data, discovery_clusters)
                    predicted_clusters = clf.predict(validation_data)

                    # Calculate stability score
                    stability_score = self.metric_fn(validation_clusters, predicted_clusters)

                    # Add additional information
                    result_entry = {
                        'split': split_id + 1,
                        'clustering_setting': setting_key,
                        self.metric: stability_score,
                        'n_clusters_discovery': len(set(discovery_clusters)),
                        'n_clusters_validation': len(set(validation_clusters))
                    }
                    return result_entry
                except Exception as e:
                    logging.warning(
                        f"Error calculating stability for split {split_id + 1}, setting {setting_key}: {str(e)}")
                    return {}

    def _get_cluster_result(self, analysis_type: str, split_id: int, setting_key: str):
        """Get clustering result for a specific analysis type, split, and setting"""
        try:
            cl_result = self.state.clustering_items[split_id]
            if hasattr(cl_result, analysis_type):
                analysis_result = getattr(cl_result, analysis_type)
                if analysis_result and setting_key in analysis_result.items:
                    return analysis_result.items[setting_key]
        except (IndexError, AttributeError, KeyError) as e:
            logging.warning(
                f"Could not retrieve clustering result for {analysis_type}, split {split_id}, setting {setting_key}: {str(e)}")
        return None

    def check_prerequisites(self):
        run_report = True

        if self.state is None or not isinstance(self.state, ClusteringState):
            logging.warning(f"{self.__class__.__name__} requires a valid state object. Report will not be created.")
            run_report = False

        return run_report



import abc
from pathlib import Path
import logging

from immuneML.data_model.EncodedData import EncodedData
from immuneML.environment.Label import Label
from immuneML.ml_methods.util.Util import Util


class MLMethod(metaclass=abc.ABCMeta):
    """
    ML method classifiers are algorithms which can be trained to predict some label on immune
    repertoires, receptors or sequences.

    These methods can be trained using the :ref:`TrainMLModel` instruction, and previously trained
    models can be applied to new data using the :ref:`MLApplication` instruction.

    When choosing which ML method(s) are most suitable for your use-case, please consider the following table:

    .. csv-table:: ML methods properties
       :file: ../../source/_static/files/ml_methods_properties.csv
       :header-rows: 1

    """
    DOCS_TITLE = "Classifiers"

    def __init__(self, name: str = None, label: Label = None):
        self.name = name
        self.label = label
        self.class_mapping = None
        self.feature_names = None
        self.optimization_metric = None
        self.result_path = None

    def fit(self, encoded_data: EncodedData, label: Label, optimization_metric: str = None, cores_for_training: int = 2):
        """
        The fit method is called by the MLMethodTrainer to fit a model to a specific encoded dataset.
        This method internally calls methods to initialise and perform model fitting.

        This method should not be overwritten.
        """
        self._initialize_fit(encoded_data, label, optimization_metric)
        self._fit(encoded_data, cores_for_training)

    def _initialize_fit(self, encoded_data: EncodedData, label: Label, optimization_metric: str):
        """
        Sets parameters and performs checks prior to fitting a model to encoded data.
        The following parameters will be set and are available to be used inside '_fit':

        - label (Label): the label for which the classifier will be created.
          Note that an individual MLMethod should handle only one label (individual methods will be trained for each provided label in TrainMLModelInstruction).

        - class_mapping (dict): This class mapping may be used to convert the original classes (strings, numbers or booleans) to numeric values.
          This can be useful for example if an internally used model only accepts numeric classes.
          It is not always necessary to use class_mapping: positive and negative classes may be directly retrieved from the Label object.

        - optimization_metric (str): The name of the optimization metric which is used to select the best model during cross-validation.
          This metric may be used internally for example for hyperparameter selection or earlystopping.

        - feature_names (list): The names of the features from the encoded data, which are exported alongside the model for transparency.

        The method furthermore checks and shows a warning if example weights are supplied for a model which cannot use
        example weights during fitting.

        This method should not be overwritten.
        """
        self.label = label
        self.class_mapping = Util.make_class_mapping(encoded_data.labels[label.name], label)
        self.optimization_metric = optimization_metric
        self.feature_names = encoded_data.feature_names

        if encoded_data.example_weights is not None and not self.can_fit_with_example_weights():
            logging.warning(f"{self.__class__.__name__}: cannot fit this classifier with example weights,"
                            f" fitting without example weights instead... Example weights will still be applied when "
                            f"computing evaluation metrics after fitting.")

    @abc.abstractmethod
    def _fit(self, encoded_data: EncodedData, cores_for_training: int = 2):
        """
        Fits the parameters of the machine learning model. This method should be implemented when adding a new MLMethod.
        Note that '_initialize_fit' is called before '_fit', such that any parameters set during initialization are
        available during fitting.

        Arguments:

            encoded_data (EncodedData): an instance of EncodedData class which includes encoded examples (repertoires, receptors or sequences), their
                labels, names of the features and other additional information. Most often, only examples and labels will be used. Examples are either a
                dense numpy matrix or a sparse matrix, where columns correspond to features and rows correspond to examples. There are a few encodings
                which make multidimensional outputs that do not follow this pattern, but they are tailored to specific ML methods which require such input
                (for instance, one hot encoding and ReceptorCNN method).

            cores_for_training (int): number of processes to be used for optional parallelization.

        Returns:

            it doesn't return anything, but fits the model parameters instead
        """
        pass

    def fit_by_cross_validation(self, encoded_data: EncodedData, label: Label, optimization_metric,
                                number_of_splits: int = 5, cores_for_training: int = 2):
        """
        See also: _fit_by_cross_validation.

        This method should not be overwritten.
        """
        self._initialize_fit(encoded_data, label, optimization_metric)
        self._fit_by_cross_validation(encoded_data, number_of_splits, cores_for_training)

    def _fit_by_cross_validation(self, encoded_data: EncodedData, number_of_splits: int, cores_for_training: int):
        """
        Like _fit, but this method first models through an additional (third) level of cross-validation.
        For most ML methods it is typically not defined. An exception is scikit-learn methods,
        see SklearnMethod for a more detailed example implementation.

        Detailed info:
        In immuneML, preprocessing, encoding and ML hyperparameters can be optimized by using nested cross-validation (see TrainMLModelInstruction for more
        details). This function is in that setting the third level of nested cross-validation which can optimize only over the model hyperparameters.
        It represents an alternative to optimizing the model hyperparameters in the TrainMLModelInstruction. Which one should be used depends on the
        use-case and specific models: models based on scikit-learn implementations come with this option by default (see SklearnMethod class), while
        custom classifiers typically do not implement this and just call fit() function and throw a warning instead.

        Arguments:

            encoded_data (EncodedData): an instance of EncodedData class which includes encoded examples (repertoires, receptors or sequences), their
                labels, names of the features and other additional information. Most often, only examples and labels will be used. Examples are either a
                dense numpy matrix or a sparse matrix, where columns correspond to features and rows correspond to examples. There are a few encodings
                which make multidimensional outputs that do not follow this pattern, but they are tailored to specific ML methods which require such input
                (for instance, one hot encoding and ReceptorCNN method).

            number_of_splits (int): number of splits for the cross-validation to be performed for selection the best hyperparameters of the ML model;
                note that if this is used in combination with nested cross-validation in TrainMLModel instruction, it can result in very few examples in
                each split depending on the orginal dataset size and the nested cross-validation setup.

            cores_for_training (int): number of processes to be used for optional parallelization.


        Returns:

            it doesn't return anything, but fits the model parameters instead

        """
        logging.warning(f"{self.__class__.__name__}: fitting by cross-validation is not implemented for this classifier: fitting one model instead...")
        self._fit(encoded_data=encoded_data, cores_for_training=cores_for_training)

    def _assert_matching_label(self, label: Label):
        """Ensures the label for prediction is matching. This method should not be overwritten."""
        assert self.label == label, f"{self.__class__.__name__}: this method is fitted for {self.label}, not {label}"

    def predict(self, encoded_data: EncodedData, label: Label):
        """Safely calls '_predict' after checking the label is matching. This method should not be overwritten."""
        self._assert_matching_label(label)
        return self._predict(encoded_data)

    def predict_proba(self, encoded_data: EncodedData, label: Label):
        """Safely calls '_predict_proba' after checking the label is matching. This method should not be overwritten."""
        self._assert_matching_label(label)
        return self._predict_proba(encoded_data)

    @abc.abstractmethod
    def _predict(self, encoded_data: EncodedData):
        """
        The predict function predicts the class for the given label across examples provided in encoded data.

        Arguments:

            encoded_data (EncodedData): an instance of EncodedData class which includes encoded examples (repertoires, receptors or sequences), their
                labels, names of the features and other additional information. Most often, only examples and labels will be used. Examples are either a
                dense numpy matrix or a sparse matrix, where columns correspond to features and rows correspond to examples. There are a few encodings
                which make multidimensional outputs that do not follow this pattern, but they are tailored to specific ML methods which require such input
                (for instance, one hot encoding and ReceptorCNN method).

        Returns:

            a dictionary where the key is the label_name and the value is a list of class predictions (one prediction per example):
            e.g., {label_name: [class1, class2, class2, class1]}

        """
        pass

    @abc.abstractmethod
    def _predict_proba(self, encoded_data: EncodedData):
        """
        The predict_proba function predicts class probabilities for the given label if the model supports probabilistic output.
        If not, it should raise a warning and return predicted classes without probabilities.

        The function will return a nested dictionary. The key(s) of the outer dictionary represent the label name(s),
        and the keys of the inner dictionary the class names of the respective label.
        The utility function py:mod:`immuneML.ml_methods.util.Util.Util.make_binary_class_mapping` may be used to
        handle mapping of class names to an internal representation for binary classification.

        Arguments:

            encoded_data (EncodedData): an object of EncodedData class where the examples attribute should be used to make predictions. examples
            attribute includes encoded examples in matrix format (numpy 2D array or a sparse matrix depending on the encoding). EncodedData object
            provided here can include labels (in model assessment scenario) or not (when predicting the class probabilities on new data which has not
            been labels), so the labels attribute of the EncodedData object should NOT be used in this function, even if it is set.

        Returns:

            a nested dictionary where the outer keys represent label names, inner keys represent class names for the respective
            label, and innermost values are 1D numpy arrays with class probabilities.
            For example for instance for label CMV where the class can be either True or False and there are
            3 examples to predict the class probabilities for: {CMV: {True: [0.2, 0.55, 0.98], False: [0.8, 0.45, 0.02]}}

        """
        pass

    @abc.abstractmethod
    def store(self, path: Path):
        """
        The store function stores the object on which it is called so that it can be imported later using load function. It typically uses pickle,
        yaml or similar modules to store the information. It can store one or multiple files.

        Arguments:

            path (Path): path to folder where to store the model

        Returns:

            it does not have a return value

        """
        pass

    @abc.abstractmethod
    def load(self, path: Path):
        """
        The load function can load the model given the folder where the same class of the model was previously stored using the store function.
        It reads in the parameters of the model and sets the values to the object attributes so that the model can be reused. For instance, this is
        used in MLApplication instruction when the previously trained model is applied on a new dataset.

        Arguments:

            path (Path): path to the folder where the model was stored using store() function

        Returns:

            it does not have a return value, but sets the attribute values of the object instead

        """
        pass

    @abc.abstractmethod
    def get_params(self, for_refitting=False) -> dict:
        """
        Returns the model parameters and their values in a readable yaml-friendly way (a dictionary consisting of ints, floats, strings, lists and dictionaries).
        This may simply be vars(self), but if an internal (sklearn) model is fitted, the parameters of the internal model should
        be included as well.

        Args:
            for_refitting:
        """
        pass

    @abc.abstractmethod
    def can_predict_proba(self) -> bool:
        """
        Returns whether the ML model can be used to predict class probabilities or class assignment only.
        This method should be overwritten to return True if probabilities can be predicted, and False if they cannot.
        """
        pass

    @abc.abstractmethod
    def can_fit_with_example_weights(self) -> bool:
        """
        Returns a boolean value indicating whether the model can be fit with example weights.
        Example weights allow to up-weight the importance of certain examples, and down-weight the importance of others.
        """
        pass

    @abc.abstractmethod
    def get_compatible_encoders(self):
        """Returns a list of compatible encoders. This method should be overwritten for every MLMethod.

        for example:

        from immuneML.encodings.evenness_profile.EvennessProfileEncoder import EvennessProfileEncoder

        return [EvennessProfileEncoder]
        """
        pass

    def check_encoder_compatibility(self, encoder):
        """
        Checks whether the given encoder is compatible with this ML method, and throws an error if it is not.

        This method should not be overwritten.
        """
        is_valid = False

        for encoder_class in self.get_compatible_encoders():
            if issubclass(encoder.__class__, encoder_class):
                is_valid = True
                break

        if not is_valid:
            raise ValueError(f"{encoder.__class__.__name__} is not compatible with ML Method {self.__class__.__name__}. "
                             f"Please use one of the following encoders instead: {', '.join([enc_class.__name__ for enc_class in self.get_compatible_encoders()])}")

    def get_package_info(self) -> str:
        """
        Should return the relevant version numbers of immuneML and all external packages that were used for the immuneML implementation.
        This information will be exported and is required for transparency and reproducibility.

        This method should be overwritten to add any additional packages if necessary. For instance, versions of scikit-learn.
        """
        return Util.get_immuneML_version()

    def get_feature_names(self) -> list:
        return self.feature_names

    def get_label_name(self) -> str:
        return self.label.name

    def get_classes(self):
        return self.label.values

    def get_positive_class(self):
        return self.label.positive_class



import argparse
import sys
from pathlib import Path
import os.path
import logging

from immuneML.api.galaxy.Util import Util
from immuneML.data_model.bnp_util import write_yaml
from immuneML.util.PathBuilder import PathBuilder


def parse_command_line_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for building specification for applying previously trained ML models in Galaxy")

    parser.add_argument("-t", "--trained_model", required=True, help="The trained ML model to apply to the dataset.")

    parser.add_argument("-o", "--output_path", required=True,
                        help="Output location for the generated yaml file (directory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")

    return parser.parse_args(args)


def build_specs(parsed_args):
    specs = {
        "definitions": {
            "datasets": {
                "dataset": {
                    "format": "AIRR",
                    "params": {"dataset_file": Util.discover_dataset_path()}
                }
            },
        },
        "instructions": {
            f"apply_ml_model": {
                "type": "MLApplication",
                "dataset": "dataset",
                "number_of_processes": 8,
                "config_path": parsed_args.trained_model
            }
        }
    }

    return specs

def main(args):
    parsed_args = parse_command_line_arguments(args)
    specs = build_specs(parsed_args)

    if not os.path.isfile(parsed_args.trained_model):
        logging.warning(f"Could not locate trained ML model: {parsed_args.trained_model}")

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    write_yaml(output_location, specs)

    return str(output_location)


if __name__ == "__main__":
    main(sys.argv[1:])



from pathlib import Path
from typing import Tuple, List

import pandas as pd
import plotly.express as px

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.ml_methods.generative_models.PWM import PWM
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.gen_model_reports.GenModelReport import GenModelReport
from immuneML.util.PathBuilder import PathBuilder


class PWMSummary(GenModelReport):
    """
    This report provides the summary of the baseline PWM and shows the following:

    - probabilities of generated sequences having different lengths
    - PWMs for each length with positive probability

    This report takes no input arguments.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        my_pwm_summary: PWMSummary

    """

    @classmethod
    def build_object(cls, **kwargs):
        name = kwargs["name"] if "name" in kwargs else "PWMSummary"
        return PWMSummary(name=name)

    def __init__(self, dataset: Dataset = None, model: GenerativeModel = None,
                 result_path: Path = None, name: str = None):
        super().__init__(dataset, model, result_path, name)
        self.info = ("This reports provides a summary of baseline PWM model and shows the probabilities of generated "
                     "sequences having different lengths, as well as logo plots for each length with positive "
                     "probability.")

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        len_figure, len_table = self._report_length_probs()
        pwm_figures, pwm_tables = self._report_pwms()
        return ReportResult(name=self.name, info=self.info, output_figures=[len_figure] + pwm_figures,
                            output_tables=[len_table] + pwm_tables)

    def _report_length_probs(self) -> Tuple[ReportOutput, ReportOutput]:
        df = pd.DataFrame({'sequence_length': self.model.length_probs.keys(), 'probability': self.model.length_probs.values()})
        df.to_csv(str(self.result_path / 'length_probs.csv'), index=False)
        len_table = ReportOutput(path=self.result_path / 'length_probs.csv', name='Sequence length probabilities')

        fig = px.bar(df, x="sequence_length", y="probability")
        fig.update_layout(xaxis=dict(tickmode='array', tickvals=df["sequence_length"]),
                          template="plotly_white")
        fig.update_traces(marker_color=px.colors.diverging.Tealrose[0])
        file_path = PlotlyUtil.write_image_to_file(fig, self.result_path / 'length_probs.html')
        len_figure = ReportOutput(path=file_path, name="Sequence length probabilities")

        return len_figure, len_table

    def _report_pwms(self) -> Tuple[List[ReportOutput], List[ReportOutput]]:
        figure_outputs = []
        table_outputs = []
        for seq_len in self.model.pwm_matrix:
            df = pd.DataFrame(data=self.model.pwm_matrix[seq_len],
                              index=EnvironmentSettings.get_sequence_alphabet(self.model.sequence_type))
            df = pd.melt(df.reset_index().rename(columns={'index': 'letter'}), id_vars=['letter'], var_name='position',
                         value_name='probability')
            figure = px.bar(df, x="position", y="probability", color="letter", text="letter",
                            color_discrete_map=PlotlyUtil.get_amino_acid_color_map(), template="plotly_white")
            figure.update_layout(showlegend=False)

            fig_path = self.result_path / f"pwm_len_{seq_len}.html"
            fig_path = PlotlyUtil.write_image_to_file(figure, fig_path)
            figure_outputs.append(ReportOutput(fig_path, f"PWM (sequence length = {seq_len})"))

            table_path = self.result_path / f"pwm_len_{seq_len}.csv"
            df.to_csv(str(table_path), index=False)
            table_outputs.append(ReportOutput(table_path, f"PWM (sequence length = {seq_len})"))

        return figure_outputs, table_outputs

    def check_prerequisites(self) -> bool:
        return isinstance(self.model, PWM)



import logging
import os
import subprocess
from collections import defaultdict
from pathlib import Path

import pandas as pd
import plotly.graph_objects as go

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class NodeDegreeDistribution(DataReport):
    """
    A report that uses CompAIRR to compute the node degree distribution of a sequence dataset. Results are visualized
    as a histogram and stored in a TSV file.

    The report assumes that CompAIRR (https://github.com/uio-bmi/compairr) has been installed.

    **Specification arguments**:

    - compairr_path (str): The path to the CompAIRR executable.

    - region_type (str): The region type to analyze. Can be either "IMGT_CDR3" or "IMGT_JUNCTION".

    - indels (bool): Whether to include indels in the analysis. Default is False.

    - ignore_genes (bool): Whether to ignore gene names in the analysis. Default is False.

    - hamming_distance (int): The Hamming distance to use for the analysis. Default is 1.

    - per_repertoire (bool): Whether to compute the node degree distribution for each repertoire separately. Only
      applicable when using a RepertoireDataset. Default is False.

    - per_label (bool): Whether to compute the node degree distribution for each label separately. Only applicable
      when using a RepertoireDataset. Default is False.

    - threads (int): The number of threads to use for the analysis. Default is 4.

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        NodeDegreeDistribution:
            compairr_path: /path/to/compairr
            region_type: IMGT_JUNCTION
            indels: False
            ignore_genes: False
            hamming_distance: 1
            per_repertoire: False
            per_label: False
            threads: 4

    """

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_region_type(kwargs)
        return NodeDegreeDistribution(**{**kwargs, 'region_type': RegionType[kwargs['region_type'].upper()]})

    def __init__(self, dataset: Dataset = None, result_path: Path = None,
                 name: str = None, compairr_path: str = None, region_type: RegionType = RegionType.IMGT_JUNCTION,
                 indels: bool = False, ignore_genes: bool = False, hamming_distance: int = 1,
                 per_repertoire: bool = False, per_label=False, threads: int = 4):

        super().__init__(dataset=dataset, result_path=result_path, name=name)
        self.compairr_path = compairr_path
        self.region_type = region_type
        self.indels = indels
        self.ignore_genes = ignore_genes
        self.hamming_distance = hamming_distance
        self.per_repertoire = per_repertoire
        self.per_label = per_label
        self.threads = threads

    def check_prerequisites(self) -> bool:
        if not self.compairr_path:
            logging.warning("CompAIRR path not provided. CompAIRR must be installed and available in the system PATH.")
            return False
        if not isinstance(self.dataset, (SequenceDataset, RepertoireDataset)):
            logging.warning(f"{NodeDegreeDistribution.__name__} report can only be generated for SequenceDataset or "
                            f"RepertoireDataset.")
            return False
        if self.region_type not in {RegionType.IMGT_CDR3, RegionType.IMGT_JUNCTION}:
            logging.warning(f"Invalid region type: {self.region_type.value.upper()}. "
                            f"{NodeDegreeDistribution.__name__} report can only be generated for "
                            f"{RegionType.IMGT_CDR3.value.upper()} or {RegionType.IMGT_JUNCTION.value.upper()}.")
            return False
        return True

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)

        output_figures, output_tables = [], []

        if isinstance(self.dataset, SequenceDataset):
            self._generate_for_sequence_dataset(output_figures, output_tables)
        elif isinstance(self.dataset, RepertoireDataset):
            self._generate_for_repertoire_dataset(output_figures, output_tables)

        return ReportResult(
            name=self.name,
            info=f"Node degree distribution with hamming distance {self.hamming_distance}",
            output_figures=output_figures,
            output_tables=output_tables
        )

    def _generate_for_sequence_dataset(self, output_figures, output_tables):
        node_degree_dist = _compute_node_degree_dist(self.result_path, self.dataset.filename, self.region_type,
                                                     self.ignore_genes, self.compairr_path, self.hamming_distance,
                                                     self.indels, self.threads)
        output_figures.append(_plot_node_degree_dist(self.result_path, node_degree_dist, "node_degree_distribution"))
        output_tables.append(_store_node_degree_dist(self.result_path, node_degree_dist, "node_degree_distribution"))

    def _generate_for_repertoire_dataset(self, output_figures, output_tables):
        node_degree_dist_list, node_degree_dist_labels = self._compute_repertoire_node_degrees()

        if self.per_repertoire:
            self._generate_output_per_repertoire(output_figures, output_tables, node_degree_dist_list)

        if self.per_label:
            self._generate_output_per_label(output_figures, output_tables, node_degree_dist_labels)

        self._generate_average_output(output_figures, output_tables, node_degree_dist_list)

    def _compute_repertoire_node_degrees(self):
        node_degree_dist_list = []
        node_degree_dist_labels = defaultdict(lambda: defaultdict(list))

        for repertoire in self.dataset.repertoires:
            node_degree_dist = _compute_node_degree_dist(
                self.result_path, repertoire.data_filename,
                self.region_type, self.ignore_genes, self.compairr_path,
                self.hamming_distance, self.indels, self.threads
            )
            node_degree_dist_list.append(node_degree_dist)

            for label in self.dataset.labels:
                node_degree_dist_labels[label][repertoire.metadata[label]].append(node_degree_dist)

        return node_degree_dist_list, node_degree_dist_labels

    def _generate_output_per_repertoire(self, output_figures, output_tables, node_degree_dist_list):
        for dist, repertoire in zip(node_degree_dist_list, self.dataset.repertoires):
            name = f"node_degree_distribution_{repertoire.metadata['subject_id']}"
            output_figures.append(_plot_node_degree_dist(self.result_path, dist, name))
            output_tables.append(_store_node_degree_dist(self.result_path, dist, name))

    def _generate_output_per_label(self, output_figures, output_tables, node_degree_dist_labels):
        for label, value_map in node_degree_dist_labels.items():
            for value, dists in value_map.items():
                name = f"node_degree_distribution_{label}_{value}"
                output_figures.append(_plot_avg_node_degree_dist(self.result_path, dists, name))
                output_tables.append(_store_avg_node_degree_dist(self.result_path, dists, name))

    def _generate_average_output(self, output_figures, output_tables, node_degree_dist_list):
        name = "node_degree_distribution_average"
        output_figures.append(_plot_avg_node_degree_dist(self.result_path, node_degree_dist_list, name))
        output_tables.append(_store_avg_node_degree_dist(self.result_path, node_degree_dist_list, name))


def _compute_node_degree_dist(result_path, dataset_filename, region_type, ignore_genes, compairr_path, hamming_distance,
                              indels, threads):
    compairr_existence_output = _run_compairr(result_path, dataset_filename, region_type, ignore_genes, compairr_path,
                                              hamming_distance, indels, threads)

    if compairr_existence_output is None:
        raise RuntimeError("CompAIRR execution failed to generate output.")

    compairr_existence_output["degree"] -= 1
    node_degree_dist = compairr_existence_output["degree"].value_counts()

    return node_degree_dist


def _run_compairr(result_path, dataset_filename, region_type, ignore_genes, compairr_path, hamming_distance, indels,
                  threads):
    output_file = result_path / f"compairr_existence.tsv"

    dedupliacted_dataset_filename = _deduplicate_sequences(dataset_filename, region_type, ignore_genes)
    cmd_args = [str(compairr_path), "--existence", str(dedupliacted_dataset_filename),
                str(dedupliacted_dataset_filename), "--output", str(output_file), "--differences",
                str(hamming_distance), "--ignore-counts"]
    if indels:
        cmd_args.append("--indels")
    if ignore_genes:
        cmd_args.append("--ignore-genes")
    if region_type == RegionType.IMGT_CDR3:
        cmd_args.append("--cdr3")

    cmd_args.extend(["-t", str(threads)])

    subprocess.run(cmd_args, capture_output=True, text=True)

    compairr_existence_output = None
    if output_file.is_file() and os.path.getsize(output_file) > 0:
        compairr_existence_output = pd.read_csv(output_file, sep='\t', names=['sequence_id', 'degree'],
                                                header=0)
        os.remove(str(output_file))

    os.remove(dedupliacted_dataset_filename)

    return compairr_existence_output


def _deduplicate_sequences(dataset_filename, region_type, ignore_genes):
    deduplicated_dataset_filename = dataset_filename.with_name(
        f"{dataset_filename.stem}_deduplicated{dataset_filename.suffix}"
    )
    dataset = pd.read_csv(dataset_filename, sep="\t", header=0)
    subset = _get_deduplication_subset(region_type, ignore_genes)

    deduplicated_dataset = dataset.drop_duplicates(subset=subset)
    deduplicated_dataset.to_csv(deduplicated_dataset_filename, sep="\t", index=False)

    return deduplicated_dataset_filename


def _get_deduplication_subset(region_type, ignore_genes):
    if region_type == RegionType.IMGT_CDR3:
        subset = ["cdr3_aa"]
    elif region_type == RegionType.IMGT_JUNCTION:
        subset = ["junction_aa"]
    else:
        raise ValueError(f"Unsupported region type: {region_type}")

    if not ignore_genes:
        subset.extend(["v_call", "j_call"])

    return subset


def _plot_node_degree_dist(result_path, node_degree_dist: pd.DataFrame, name: str) -> ReportOutput:
    fig = go.Figure()
    fig.add_trace(go.Bar(x=node_degree_dist.index, y=node_degree_dist.values))
    fig.update_layout(xaxis_title="Degree",
                      yaxis_title="Count")
    output_path = result_path / f"{name}.html"
    output_path = PlotlyUtil.write_image_to_file(fig, output_path)

    return ReportOutput(path=output_path, name=name)


def _plot_avg_node_degree_dist(result_path, node_degree_dists: list[pd.Series], name: str) -> ReportOutput:
    df = pd.DataFrame(node_degree_dists).fillna(0).astype(int)
    avg_dist = df.mean(axis=0)
    std_dist = df.std(axis=0)

    fig = go.Figure()
    fig.add_trace(go.Bar(
        x=avg_dist.index,
        y=avg_dist.values,
        error_y=dict(type='data', array=std_dist.values, visible=True)
    ))

    fig.update_layout(xaxis_title="Degree", yaxis_title="Average Count ± Std Dev")
    output_path = result_path / f"{name}.html"
    output_path = PlotlyUtil.write_image_to_file(fig, output_path)

    return ReportOutput(path=output_path, name=f"{name}")


def _store_node_degree_dist(result_path, node_degree_dist: pd.Series, name: str) -> ReportOutput:
    output_path = result_path / f"{name}.tsv"
    node_degree_dist.to_csv(output_path, sep="\t")
    return ReportOutput(path=output_path, name=name)


def _store_avg_node_degree_dist(result_path, node_degree_dists: list[pd.Series], name: str) -> ReportOutput:
    df = pd.DataFrame(node_degree_dists).fillna(0).astype(int)
    avg_dist = df.mean(axis=0)
    std_dist = df.std(axis=0)

    combined = pd.DataFrame({
        "degree": avg_dist.index,
        "average_count": avg_dist.values,
        "std_count": std_dist.values
    })

    output_path = result_path / f"{name}.tsv"
    combined.to_csv(output_path, sep="\t", index=False)

    return ReportOutput(path=output_path, name=f"{name}")
